import inspect
import os
import sys
import tempfile
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from biocore import DataHandler
from biocore.utils.import_util import is_ipywidgets_available, is_matplotlib_available
from biocore.utils.naming import camelcase_to_snakecase
from biocore.utils.py_util import is_dataset_dict

from biofit.integration.R.r_caller import RCaller
from biofit.processing import (
    _ORDERED_FORMATS,
    TransformationMixin,
    sync_backup_config,
)
from biofit.utils import (
    _build_cache_dir,
    fingerprint_from_data,
    has_ext,
    has_separator,
    logging,
    move_temp_file,
)
from biofit.visualization.plotting_utils import (
    display_image_carousel,
    is_in_notebook,
)

from ..processing import BaseConfig, SelectedColumnTypes, SelectedFeatureTypes

logger = logging.get_logger(__name__)


ORDERED_PLOTTER_FORMATS = _ORDERED_FORMATS + ["dataset", "ds"]


def _processor_info_from_fingerprint(fingerprint: str):
    if fingerprint is None:
        return "", "", ""
    processor_info = fingerprint.split("-")
    ds_name = ""
    if len(processor_info) == 5:
        _, processor_name, processor_type, ds_name, _ = processor_info
    elif len(processor_info) == 4:
        _, processor_name, processor_type, ds_name = processor_info
    elif len(processor_info) == 3:
        _, processor_name, processor_type = processor_info
    else:
        return "", "", ""
    return processor_name, processor_type, ds_name


@dataclass
class PlotterConfig(BaseConfig):
    path: str = field(default=None, init=True, repr=True)
    device: str = field(default="pdf", init=True, repr=False)
    fingerprint: str = field(default=None, init=True, repr=False)
    unused_columns: SelectedColumnTypes = field(default=None, init=True, repr=False)
    raise_if_missing: bool = field(default=True, init=True, repr=False)
    cache_dir: str = field(default=None, init=True, repr=False)
    version: str = field(default="0.0.0", init=True, repr=False)

    _input_columns: SelectedColumnTypes = field(default=None, init=False, repr=False)
    _compare: bool = field(default=False, init=False, repr=False)
    _fit_input_feature_types: SelectedFeatureTypes = field(
        default=None, init=False, repr=False
    )
    _fit_unused_feature_types: SelectedFeatureTypes = field(
        default=None, init=False, repr=False
    )
    _transform_input_feature_types: SelectedFeatureTypes = field(
        default=None, init=False, repr=False
    )
    _transform_unused_feature_types: SelectedFeatureTypes = field(
        default=None, init=False, repr=False
    )
    r_code: str = field(default=None, init=False, repr=False)
    r_source: str = field(default=None, init=False, repr=False)
    main_method: str = field(default=None, init=False, repr=False)

    processor_type: str = field(default="", init=False, repr=False)
    processor_name: str = field(default="", init=False, repr=False)
    dataset_name: str = field(default="", init=False, repr=False)

    # automatically generated attributes

    feature_idx_in_: List[int] = field(default=None, init=False, repr=False)
    feature_names_in_: List[str] = field(default=None, init=False, repr=False)
    extra_idx_in_: List[List[int]] = field(default=None, init=False, repr=False)
    extra_names_in_: List[List[str]] = field(default=None, init=False, repr=False)


class BasePlotter(TransformationMixin):
    """_summary_

    Attributes:
        r_code (_type_): _description_
        r_source (_type_): _description_
        feature_type (_type_): _description_
        dtype (str): specifies the type of the plotter. Can be 'plotter' or 'plotter_for'

    Raises:
        NotImplementedError: _description_
        ValueError: _description_
        ValueError: _description_
        ValueError: _description_
        ValueError: _description_
        ValueError: _description_
        ValueError: _description_
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    r_caller: RCaller = None
    _config_class = PlotterConfig
    plotter: Optional[str] = None
    config: PlotterConfig

    def __init__(self, config: Optional[PlotterConfig] = None, **kwargs):
        add_new_attr = kwargs.pop("add_new_attr", False)
        ignore_none = kwargs.pop("ignore_none", False)

        if config is None:
            if hasattr(self, "_config_class"):
                self.config = self._config_class.from_dict(
                    kwargs, ignore_none=ignore_none, add_new_attr=add_new_attr
                )
        elif isinstance(config, PlotterConfig):
            self.config = config
        elif isinstance(config, dict):
            self.config = self._config_class.from_dict(
                config, ignore_none=ignore_none, add_new_attr=add_new_attr
            )
        else:
            raise ValueError(f"Unsupported config type {type(config)}")
        if config is None:
            self = self.set_params(**kwargs)
        if self.config.r_source and self.config.main_method and self.plotter is None:
            self.r_caller = RCaller.from_script(self.config.r_source)
            self.r_caller.verify_r_dependencies(
                install_missing=kwargs.get("install_missing")
            )
            self.plotter = self.r_caller.get_method(self.config.main_method)
        if kwargs.get("_function", None):
            self._function = kwargs["_function"]

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            self.plotter = self.r_caller.get_method(self.config.main_method)

        return self

    @classmethod
    def _from_config(cls, config: PlotterConfig, **kwargs):
        return cls(config=config, **kwargs)

    def plot(self, X, *args, **kwargs):
        if is_dataset_dict(X):
            raise ValueError(
                "Cannot plot a DatasetDict or IterableDatasetDict. Please provide a Dataset or IterableDataset."
            )
        return self._plot(X, *args, **kwargs)

    def _plot(self, X, *args, **kwargs):
        """
        Transforms the input data.

        Args:
            X (Any): The input data.
            *args: Additional arguments. These are additional table or array-like data to be used in the plotting.
            **kwargs: Additional keyword arguments. These are objects or parameters to be used in the plotting.

        Returns:
            Any: The computed processor.
        """
        show = kwargs.pop("show", True)
        output_dir = kwargs.pop("output_dir", None)
        self.config, kwargs = self.config.replace_defaults(
            ignore_none=True, return_unused_kwargs=True, **kwargs
        )

        args = list(args)

        plot_funcs = self._get_method(ORDERED_PLOTTER_FORMATS, "plot")
        plot_func, target_format = self._get_target_func(
            plot_funcs,
            source_format=DataHandler.get_format(X),
            target_formats=kwargs.get("input_format", None),
            accepted_formats=ORDERED_PLOTTER_FORMATS,
        )

        fingerprint = kwargs.pop("fingerprint", None) or self.config.fingerprint

        image_paths = None
        if plot_func:
            (
                self.feature_names_in_,
                self.feature_idx_in_,
                _,
                self.extra_names_in_,
                self.extra_idx_in_,
                _,
                _,
            ) = self._get_columns(
                X,
                *args,
                input_columns=self.config._input_columns,
                unused_columns=self.config.unused_columns,
                input_feature_types=self.config._transform_input_feature_types,
                unused_feature_types=self.config._transform_unused_feature_types,
                raise_if_missing=self.config.raise_if_missing,
            )

            path_to_save = self.config.path

            # required args are ones without defaults
            required_args = [
                p.default == p.empty
                for p in inspect.signature(plot_func).parameters.values()
            ][1:]
            if (
                not self.config._compare
                and self.extra_idx_in_ is not None
                and len(self.extra_idx_in_) > 0
                and DataHandler.supports_named_columns(X)
            ):
                new_cols = self._make_columns_exclusive(
                    [self.feature_names_in_] + self.extra_names_in_
                )

                self.feature_names_in_ = new_cols[0]
                self.feature_idx_in_ = sorted(
                    DataHandler.get_column_indices(X, self.feature_names_in_)
                )

                self.extra_names_in_ = new_cols[1:]
                extra_idx_in_ = []
                for i, names in enumerate(self.extra_names_in_):
                    if len(args) and args[i] is not None:
                        extra_idx_in_.append(
                            sorted(
                                DataHandler.get_column_indices(
                                    args[i], names, raise_if_missing=False
                                )
                            )
                        )
                    elif names is not None:
                        extra_idx_in_.append(
                            sorted(
                                DataHandler.get_column_indices(
                                    X, names, raise_if_missing=True
                                )
                            )
                        )
                    else:
                        if required_args[i]:
                            raise ValueError(
                                f"Missing required argument {i} for plot function {plot_func}"
                            )
                        extra_idx_in_.append(self.extra_idx_in_[i])
                self.extra_idx_in_ = extra_idx_in_

            data_fingerprint = fingerprint_from_data(X)
            if fingerprint is None:
                fingerprint = self.generate_fingerprint(data_fingerprint, self.config)

            input, args, kwargs = self._process_plot_input(X, *args, **kwargs)
            args = list(args)
            if len(args) == 0:
                if self.extra_idx_in_:
                    for i, inds in enumerate(self.extra_idx_in_):
                        if inds is not None:
                            args.append(
                                DataHandler.to_format(
                                    X, target_format, input_columns=inds
                                )
                            )
                        else:
                            args.append(None)
            else:
                for i, (inds, arg) in enumerate(zip(self.extra_idx_in_, args)):
                    if arg is not None:
                        args[i] = DataHandler.to_format(
                            arg, target_format, input_columns=inds
                        )
                    elif inds is not None:
                        args[i] = DataHandler.to_format(
                            X, target_format, input_columns=inds
                        )

            input = DataHandler.to_format(
                input, target_format, input_columns=self.feature_idx_in_
            )
            temp_dir = tempfile.mkdtemp()
            temp_file = os.path.join(temp_dir, fingerprint)
            temp_dir = Path(temp_dir)
            temp_dir.mkdir(parents=True, exist_ok=True)
            temp_dir = temp_dir.resolve().as_posix()
            self.config.path = temp_file + f".{self.config.device}"
            output_dir = None if output_dir is None else Path(output_dir)
            file_name = None
            if path_to_save:
                if has_separator(path_to_save):
                    if has_ext(path_to_save):
                        output_dir = (
                            Path(path_to_save).parent
                            if output_dir is None
                            else output_dir
                        )
                        file_name = Path(path_to_save).name
                    else:
                        output_dir = (
                            Path(path_to_save) if output_dir is None else output_dir
                        )
                        file_name = None
                elif has_ext(path_to_save):
                    file_name = str(path_to_save)

            if output_dir is None:
                if kwargs.get("processor", None) and kwargs["processor"].cache_files:
                    output_dir = (
                        Path(kwargs["processor"].cache_files[0]["filename"]).parent
                        / "plots"
                    )
                elif hasattr(X, "cache_files") and X.cache_files:
                    output_dir = Path(X.cache_files[0]["filename"]).parent / "plots"
                else:
                    output_dir = Path(_build_cache_dir(X, data_fingerprint)) / "plots"
            if file_name is None:
                cls_name = camelcase_to_snakecase(
                    self.__class__.__name__.replace("_plotter", "")
                )
                file_name = f"{fingerprint}_{cls_name}.{self.config.device}"

            fig = plot_func(input, *args, **kwargs)
            if "matplotlib" in sys.modules:
                import matplotlib.pyplot as plt

            if output_dir:

                def get_or_move_images(images, output_dir, move_images=False):
                    image_paths = []
                    if len(images) > 1:
                        for fn in images:
                            old_name = fn.resolve().as_posix()
                            new_name = f"{output_dir}/{fn.name}"
                            if move_images:
                                image_paths.append(new_name)
                                move_temp_file(old_name, new_name)
                            else:
                                image_paths.append(old_name)
                        if move_images:
                            logger.info(f"Saved {len(images)} plots to {output_dir}")
                    elif len(images) == 1:
                        old_name = images[0].resolve().as_posix()
                        if move_images:
                            image_paths = f"{output_dir}/{file_name}"
                            move_temp_file(old_name, image_paths)
                            logger.info(f"Saved plot to {image_paths}")
                        else:
                            image_paths = old_name

                    return image_paths

                output_dir.mkdir(parents=True, exist_ok=True)
                output_dir = output_dir.resolve().as_posix()
                # move all files within temp_dir to output_dir
                images = [fn for fn in Path(temp_dir).glob(f"*.{self.config.device}")]
                image_paths = get_or_move_images(images, output_dir, move_images=True)

                _image_paths = []
                if self.config.device != "png":
                    _images = [fn for fn in Path(temp_dir).glob("*.png")]
                    _image_paths = get_or_move_images(
                        _images, output_dir, move_images=False
                    )
                if len(_image_paths) > 0:
                    image_paths = _image_paths
                    images = _images

                if len(image_paths) > 1:
                    if is_in_notebook() and show and is_ipywidgets_available():
                        display_image_carousel(image_paths, "png")
                elif len(image_paths) == 1:
                    if is_in_notebook() and show:
                        if (
                            is_matplotlib_available()
                            and "matplotlib" in sys.modules
                            and isinstance(fig, plt.Figure)
                        ):
                            # ignore warning about non-interactive backend
                            warnings.filterwarnings("ignore")
                            fig.show()
                            warnings.filterwarnings("default")
                        else:
                            from IPython.display import Image, display

                            display(
                                Image(
                                    image_paths,
                                    embed=True,
                                    format="png",
                                    width=720,
                                )
                            )
                    elif (
                        is_matplotlib_available()
                        and "matplotlib" in sys.modules
                        and isinstance(fig, plt.Figure)
                    ):
                        plt.close(fig)
                else:
                    logger.warning("No plots were generated")
        return image_paths

    def _process_plot_input(self, input, *args, **kwargs):
        return input, args, kwargs



import textwrap
from dataclasses import dataclass, field
from typing import List, Type

import biofit.config as config
from biofit.integration.biosets import get_feature
from biofit.integration.R import RCaller
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils.types import Unset
from biofit.visualization.plotting import BasePlotter, PlotterConfig


@dataclass
class ViolinConfig(PlotterConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )

    r_source: str = field(
        default=(config.R_SCRIPTS / "plotting_utils.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="generate_violin", init=False, repr=False)

    column: str = None
    label_name: str = "labels"
    xlab: str = "Labels"
    ylab: str = "Value"


class ViolinPlotter(BasePlotter):
    _config_class = ViolinConfig
    config: ViolinConfig

    def __init__(
        self,
        column: str = None,
        label_name: str = None,
        xlab: str = Unset('"Labels"'),
        ylab: str = Unset('"Value"'),
        install_missing: bool = None,
        config=None,
    ):
        super().__init__(
            config=config, xlab=xlab, ylab=ylab, install_missing=install_missing
        )
        self.plotter = None
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

        return self

    def plot(
        self,
        x,
        y=None,
        column: SelectedColumnTypes = None,
        label_name: SelectedColumnTypes = None,
        xlab: str = Unset('"Labels"'),
        ylab: str = Unset('"Value"'),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(
            column, label_name
        )
        return self._plot(
            x,
            y,
            xlab=xlab,
            ylab=ylab,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_arrow(self, x, y=None):
        self.plotter(
            x,
            y,
            **self.config.get_params(),
        )



from dataclasses import dataclass, field

from biofit.processing import BaseProcessor, ProcessorConfig


@dataclass
class ImputationConfig(ProcessorConfig):
    """Base class for imputation processor configurations."""

    processor_name: str = field(default="imputation", init=False, repr=False)


class Imputation(BaseProcessor):
    """Base class for imputation processors."""



import importlib
from collections import OrderedDict
from typing import TYPE_CHECKING

from biocore.utils.import_util import is_transformers_available

from biofit.utils import logging

from ..auto.configuration_auto import (
    PROCESSOR_CATEGORY_MAPPING_NAMES,
    PROCESSOR_TYPE_MAPPING_NAMES,
    AutoConfig,
)

if TYPE_CHECKING:
    from biofit.processing import BaseProcessor

logger = logging.get_logger(__name__)


def _get_model_class(config, model_mapping: "_LazyAutoMapping"):
    supported_models = model_mapping[type(config)]
    if not isinstance(supported_models, (list, tuple)):
        return supported_models

    name_to_model = {model.__name__: model for model in supported_models}
    architectures = getattr(config, "architectures", [])
    for arch in architectures:
        if arch in name_to_model:
            return name_to_model[arch]
        elif f"TF{arch}" in name_to_model:
            return name_to_model[f"TF{arch}"]
        elif f"Flax{arch}" in name_to_model:
            return name_to_model[f"Flax{arch}"]

    # If not architecture is set in the config or match the supported models, the first element of the tuple is the
    # defaults.
    return supported_models[0]


def _get_class(config, model_mapping):
    supported_models = model_mapping[type(config)]
    return supported_models


class _BaseAutoProcessorClass:
    # Base class for auto preprocessors.
    _processor_mapping = None

    def __init__(self, *args, **kwargs):
        raise EnvironmentError(
            f"{self.__class__.__name__} is designed to be instantiated "
            f"`{self.__class__.__name__}.from_config(config)` methods."
        )

    @classmethod
    def for_processor(self, processor_name, **kwargs):
        config = AutoConfig.for_processor(processor_name, **kwargs)
        return self.from_config(config, **kwargs)

    @classmethod
    def from_config(cls, config, **kwargs):
        if type(config) in cls._processor_mapping.keys():
            preprocessor_class: "BaseProcessor" = _get_class(
                config, cls._processor_mapping
            )
            return preprocessor_class._from_config(config, **kwargs)

        raise ValueError(
            f"Unrecognized configuration class {config.__class__} for this kind of AutoProcessor: {cls.__name__}.\n"
            f"Processor type should be one of {', '.join(c.__name__ for c in cls._processor_mapping.keys())}."
        )

    @classmethod
    def register(cls, _config_class, processor_class: "BaseProcessor", exist_ok=False):
        """
        Register a new model for this class.

        Args:
            _config_class ([`PretrainedConfig`]):
                The configuration corresponding to the model to register.
            model_class ([`PreTrainedModel`]):
                The model to register.
        """
        if hasattr(processor_class, "_config_class") and issubclass(
            _config_class, processor_class._config_class
        ):
            raise ValueError(
                "The model class you are passing has a `_config_class` attribute that is not consistent with the "
                f"config class you passed (model has {processor_class._config_class} and you passed {_config_class}. Fix "
                "one of those so they match!"
            )
        cls._processor_mapping.register(
            _config_class, processor_class, exist_ok=exist_ok
        )


class _BaseAutoModelClass:
    # Base class for auto models.
    _processor_mapping = None

    def __init__(self, *args, **kwargs):
        raise EnvironmentError(
            f"{self.__class__.__name__} is designed to be instantiated "
            f"`{self.__class__.__name__}.from_config(config)` methods."
        )

    @classmethod
    def for_model(cls, model_name, *model_args, **kwargs):
        config = AutoConfig.for_processor(model_name, **kwargs)
        return cls.from_config(config, *model_args, **kwargs)

    @classmethod
    def from_config(cls, config, **kwargs):
        if type(config) in cls._processor_mapping.keys():
            model_class: "BaseProcessor" = _get_model_class(
                config, cls._processor_mapping
            )
            return model_class._from_config(config, **kwargs)

        if is_transformers_available():
            logger.warning(
                "Could not find a matching model for this configuration. "
                "Searching for a model in the Transformers library instead."
            )

            from transformers.models.auto.auto_factory import (
                _BaseAutoModelClass as _HfBaseAutoModelClass,
            )

            return _HfBaseAutoModelClass.from_config(config, **kwargs)

        raise ValueError(
            f"Unrecognized configuration class {config.__class__} for this kind of AutoModel: {cls.__name__}.\n"
            f"Model type should be one of {', '.join(c.__name__ for c in cls._processor_mapping.keys())}."
        )

    @classmethod
    def from_pretrained(cls, pretrained_estimator_name_or_path, *model_args, **kwargs):
        if is_transformers_available():
            from transformers.models.auto.auto_factory import (
                _BaseAutoModelClass as _HfBaseAutoModelClass,
            )

            return _HfBaseAutoModelClass.from_pretrained(
                pretrained_estimator_name_or_path, *model_args, **kwargs
            )
        else:
            raise EnvironmentError(
                f"Using {cls.__name__}.from_pretrained requires the transformers library to be installed. "
                "You can install it with `pip install transformers`"
            )

    @classmethod
    def register(cls, _config_class, processor_class: "BaseProcessor", exist_ok=False):
        """
        Register a new model for this class.

        Args:
            _config_class ([`PretrainedConfig`]):
                The configuration corresponding to the model to register.
            model_class ([`PreTrainedModel`]):
                The model to register.
        """
        if hasattr(processor_class, "_config_class") and issubclass(
            _config_class, processor_class._config_class
        ):
            raise ValueError(
                "The model class you are passing has a `_config_class` attribute that is not consistent with the "
                f"config class you passed (model has {processor_class._config_class} and you passed {_config_class}. Fix "
                "one of those so they match!"
            )
        cls._processor_mapping.register(
            _config_class, processor_class, exist_ok=exist_ok
        )


def insert_head_doc(docstring, head_doc=""):
    if len(head_doc) > 0:
        return docstring.replace(
            "one of the model classes of the library ",
            f"one of the model classes of the library (with a {head_doc} head) ",
        )
    return docstring.replace(
        "one of the model classes of the library ",
        "one of the base model classes of the library ",
    )


def get_values(model_mapping):
    result = []
    for model in model_mapping.values():
        if isinstance(model, (list, tuple)):
            result += list(model)
        else:
            result.append(model)

    return result


def getattribute_from_module(module, attr):
    if attr is None:
        return None
    if isinstance(attr, tuple):
        return tuple(getattribute_from_module(module, a) for a in attr)
    if hasattr(module, attr):
        return getattr(module, attr)
    # Some of the mappings have entries estimator_type -> object of another model type. In that case we try to grab the
    # object at the top level.
    biofit_module = importlib.import_module("biofit")

    if module != biofit_module:
        try:
            return getattribute_from_module(biofit_module, attr)
        except ValueError:
            raise ValueError(
                f"Could not find {attr} neither in {module} nor in {biofit_module}!"
            )
    else:
        raise ValueError(f"Could not find {attr} in {biofit_module}!")


class _LazyAutoMapping(OrderedDict):
    """
    " A mapping config to object (model or tokenizer for instance) that will load keys and values when it is accessed.

    Args:
        - config_mapping: The map model type to config class
        - model_mapping: The map model type to model (or tokenizer) class
    """

    def __init__(self, config_mapping, processor_mapping):
        self._config_mapping = config_mapping
        self._reverse_config_mapping = {v: k for k, v in config_mapping.items()}
        self._processor_mapping = processor_mapping
        self._processor_mapping._processor_mapping = self
        self._extra_content = {}
        self._modules = {}

    def __len__(self):
        common_keys = set(self._config_mapping.keys()).intersection(
            self._processor_mapping.keys()
        )
        return len(common_keys) + len(self._extra_content)

    def __getitem__(self, key):
        if key in self._extra_content:
            return self._extra_content[key]
        processor_type = self._reverse_config_mapping[key.__name__]
        if processor_type in self._processor_mapping:
            processor_name = self._processor_mapping[processor_type]
            return self._load_attr_from_module(processor_type, processor_name)

        # Maybe there was several model types associated with this config.
        estimator_types = [
            k for k, v in self._config_mapping.items() if v == key.__name__
        ]
        for ptype in estimator_types:
            if ptype in self._processor_mapping:
                processor_name = self._processor_mapping[ptype]
                return self._load_attr_from_module(ptype, processor_name)
        raise KeyError(key)

    def _load_attr_from_module(self, module_name, attr):
        if module_name not in self._modules:
            processor_category = PROCESSOR_CATEGORY_MAPPING_NAMES.get(
                module_name, "models"
            )
            processor_type = PROCESSOR_TYPE_MAPPING_NAMES.get(module_name, None)
            if processor_type is not None:
                package_name = f"biofit.{processor_category}.{processor_type}"
            else:
                package_name = f"biofit.{processor_category}"
            self._modules[module_name] = importlib.import_module(
                f".{module_name}", package_name
            )
        return getattribute_from_module(self._modules[module_name], attr)

    def keys(self):
        mapping_keys = [
            self._load_attr_from_module(key, name)
            for key, name in self._config_mapping.items()
            if key in self._processor_mapping.keys()
        ]
        return mapping_keys + list(self._extra_content.keys())

    def get(self, key, default):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __bool__(self):
        return bool(self.keys())

    def values(self):
        mapping_values = [
            self._load_attr_from_module(key, name)
            for key, name in self._processor_mapping.items()
            if key in self._config_mapping.keys()
        ]
        return mapping_values + list(self._extra_content.values())

    def items(self):
        mapping_items = [
            (
                self._load_attr_from_module(key, self._config_mapping[key]),
                self._load_attr_from_module(key, self._processor_mapping[key]),
            )
            for key in self._processor_mapping.keys()
            if key in self._config_mapping.keys()
        ]
        return mapping_items + list(self._extra_content.items())

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, item):
        if item in self._extra_content:
            return True
        if (
            not hasattr(item, "__name__")
            or item.__name__ not in self._reverse_config_mapping
        ):
            return False
        estimator_type = self._reverse_config_mapping[item.__name__]
        return estimator_type in self._processor_mapping

    def register(self, key, value, exist_ok=False):
        """
        Register a new processor in this mapping.
        """
        if hasattr(key, "__name__") and key.__name__ in self._reverse_config_mapping:
            estimator_type = self._reverse_config_mapping[key.__name__]
            if estimator_type in self._processor_mapping.keys() and not exist_ok:
                raise ValueError(f"'{key}' is already used by a Transformers model.")

        self._extra_content[key] = value



from dataclasses import dataclass, field
from typing import List, Type

import numpy as np
from scipy.interpolate import interp1d

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..scaling import Scaler, ScalerConfig

logger = logging.get_logger(__name__)


def css_percentile_fast(mat, relative_threshold=0.1):
    """Calculates the percentile for which to sum counts up to and scale by.

    Args:

    """
    # Check for sample with one or zero features
    if np.any(np.sum(mat > 0, axis=1) <= 1):
        raise ValueError("Warning: sample with one or zero features")

    mat = mat.astype(float)
    # Sort each row with elements greater than zero
    smat = np.sort(mat * (mat > 0), axis=1)[:, ::-1]
    smat[smat == 0] = np.nan

    # Pad with NAs to make all rows the same length
    max_length = np.max(np.nansum(smat > 0, axis=1))
    padded_smat = np.full((smat.shape[0], max_length), np.nan)
    for i in range(smat.shape[0]):
        valid_values = smat[i, smat[i] > 0]
        padded_smat[i, : len(valid_values)] = valid_values

    # Calculate quantiles for each row
    quantiles = np.nanquantile(
        padded_smat, np.linspace(0, 1, padded_smat.shape[1]), axis=1
    )

    padded_smat[np.isnan(padded_smat)] = 0

    # Compute column means, ignoring NaNs
    ref1 = np.nanmean(padded_smat, axis=0)[::-1]

    # Calculate differences
    diffr = ref1[:, np.newaxis] - quantiles

    # Calculate median of absolute differences
    diffr1 = np.nanmedian(np.abs(diffr), axis=1)

    diffr1[diffr1 == 0] = np.nan

    # Determine threshold
    rel_diff = np.abs(np.diff(diffr1)) / diffr1[:-1]

    x = (np.where(rel_diff > relative_threshold)[0][0] + 1) / len(diffr1)
    if x <= 0.5:
        logger.info(
            f"Percentile calculated at {x}, which is less than 0.5. Using default value instead."
        )
        x = 0.5
    return x


def css_percentile(mat: np.ndarray, approx=False, relative_threshold=0.1):
    if np.any(np.sum(mat, axis=1) == 0):
        raise ValueError("Warning: empty feature", mat)

    # Sorting each row
    smat = np.sort(mat, axis=1)
    ref = np.mean(smat, axis=0)
    orig_dtype = mat.dtype
    mat = mat.astype(float) if not np.issubdtype(orig_dtype, np.floating) else mat

    if not mat.flags["WRITEABLE"]:
        # copy the array if it is not writeable
        mat = np.array(mat)
    mat[mat == 0] = np.nan

    refS = np.sort(ref)

    k = np.where(refS > 0)[0][0]
    lo = len(refS) - k

    if not approx:
        diffr = np.apply_along_axis(
            lambda row: refS[k:] - np.nanquantile(row, np.linspace(0, 1, lo), axis=0),
            axis=1,
            arr=mat,
        )
    else:

        def f(row):
            srow = np.sort(row)
            rrow = (srow[0], np.nanmax(srow))
            y = np.arange(len(row))
            return refS[k:] - interp1d(row, y)(np.linspace(*rrow, lo))

        diffr = np.apply_along_axis(f, axis=1, arr=mat)

    mat[np.isnan(mat)] = 0
    mat = mat.astype(orig_dtype) if not np.issubdtype(orig_dtype, np.floating) else mat

    diffr2 = np.nanmedian(np.abs(diffr), axis=0)

    rel_diff = np.abs(np.diff(diffr2)) / diffr2[:-1]
    if len(rel_diff) == 0:
        return 0.5
    x = (np.where(rel_diff > relative_threshold)[0][0] + 1) / len(diffr2)
    if x <= 0.5:
        logger.info(
            f"Percentile calculated at {x}, which is less than 0.5. Using default value instead."
        )
        x = 0.5

    return x


@dataclass
class CumulativeSumScalerConfig(ScalerConfig):
    """
    Configuration for CumulativeSumScaler.
    """

    processor_name: str = field(default="css", init=False, repr=False)
    _fit_process_desc: str = field(
        default="Calculating cumulative sum scaling percentile", init=False, repr=False
    )
    _transform_process_desc: str = field(
        default="Applying cumulative sum scaling", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    scale: int = 1000
    relative_threshold: float = 0.5
    approx: bool = False
    percentile: float = None


@dataclass
class CumulativeSumScalerConfigForMetagenomics(CumulativeSumScalerConfig):
    """
    CumulativeSumScaler specifically designed for metagenomics data.
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("ReadCount"), get_feature("Abundance"))],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [(get_feature("ReadCount"), get_feature("Abundance"))],
        init=False,
        repr=False,
    )
    dataset_name: str = field(default="metagenomics", init=False, repr=False)


@dataclass
class CumulativeSumScalerConfigForOTU(CumulativeSumScalerConfig):
    """
    CumulativeSumScaler specifically designed for otu abundance.

    Args:
    """

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)


class CumulativeSumScaler(Scaler):
    """
    BaseCumSumScaler applies cumulative sum scaling to the input data.
    """

    output_dtype = "float64"
    config_class = CumulativeSumScalerConfig
    config: CumulativeSumScalerConfig

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.percentile is not None:
            # there is overhead when passing datasets to the fit method
            # (e.g. converting one format to another, preparing mapping jobs, etc.)
            # which is unnecessary if the percentile is already known
            # Therefore, we remove the fit method to avoid it being called
            # TODO: This is a temporary solution. We should find a better way to handle this
            delattr(self, "fit_numpy")
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "CumulativeSumScaler":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_numpy(self, X: np.ndarray):
        # The fast implementation of CSS requires counts of all samples to at least
        # have two non zero features.
        if np.any(np.sum(X > 0, axis=1) <= 1):
            self.config.percentile = css_percentile(X)
        else:
            self.config.percentile = css_percentile_fast(X)
        return self

    def _transform_numpy(self, X: np.ndarray):
        xx = np.where(X == 0, np.nan, X)

        # Compute row quantiles
        qs = np.nanquantile(xx, self.config.percentile, axis=1, keepdims=True)

        xx_adj = xx - np.finfo(float).eps
        norm_factors = np.nansum(np.where(xx_adj <= qs, xx_adj, 0), axis=1)
        norm_factors[norm_factors == 0] = np.nan
        norm_factors = norm_factors / self.config.scale

        return X / norm_factors[:, np.newaxis]



"""
Correlation calculation
"""

from dataclasses import dataclass, field

from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import (
    SelectedColumnTypes,
    SelectedFeatureTypes,
    sync_backup_config,
)
from biofit.utils import logging

from ..stat import Stat, StatConfig

logger = logging.get_logger(__name__)

CORRELATION_STAT_DOCSTRING = """
Stat features based on the correlation with the target variable.

Args:
    X: Input data
    y: Target variable
    input_columns: Columns to filter
    target_column: Target column
    **kwargs: Additional keyword arguments

Returns:
    SampleFiltered data.
"""


@dataclass
class CorrelationStatConfig(StatConfig):
    _transform_input_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    processor_name: str = field(default="correlation", init=False, repr=False)

    input_columns: str = None
    target_column: str = None
    method: str = "auto"

    def __post_init__(self):
        if self.method not in [
            "auto",
            "pearsonr",
            "spearmanr",
            "kendalltau",
            "pointbiserialr",
        ]:
            raise ValueError(
                f"Method {self.method} not supported. Supported methods are: pearsonr, spearmanr, kendalltau, pointbiserialr"
            )
        if self.method != "auto":
            self._transform_process_desc = (
                f"Calculating {self.method} correlation with target variable"
            )


class CorrelationStat(Stat):
    """
    Correlation calculation based on the correlation with the target variable.
    """

    config_class = CorrelationStatConfig
    config: CorrelationStatConfig
    output_dtype = "float64"
    func = None

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        import scipy.stats

        if self.config.method != "auto":
            self.func = getattr(scipy.stats, self.config.method)
        return self

    def transform(
        self,
        X,
        y,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column
        )
        return self._process_transform(
            X,
            y,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            y,
            input_columns=input_columns,
            target_column=target_column,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _check_data(self, X, y):
        if self.config.method == "pointbiserialr":
            dtypes = next(iter(DataHandler.get_dtypes(y).values()))
            if not any(dtype in dtypes for dtype in ["int", "bool", "str"]):
                raise ValueError(
                    f"Point biserial correlation can only be used with binary target variables. Data type: {dtypes}"
                )
            n_unique = DataHandler.nunique(
                y, DataHandler.get_column_names(y, generate_cols=True)[0]
            )
            if n_unique != 2:
                raise ValueError(
                    f"Point biserial correlation can only be used with binary target variables. Number of unique values: {n_unique}"
                )
            return y, X  # the first argument must be the binary target variable
        return X, y

    def _process_transform_input(self, X, **kwargs):
        if self.config.method == "auto":
            import scipy.stats

            inds = kwargs["fn_kwargs"]["extra_indices"][0]
            is_categorical = DataHandler.is_categorical(X, inds, threshold=0.05)
            if is_categorical:
                n_classes = DataHandler.nunique(X, inds)
                if n_classes == 2:
                    self.config.method = "pointbiserialr"
                    kwargs["desc"] = (
                        "Calculating point biserial correlation with target variable"
                    )
                else:
                    self.config.method = "pearsonr"
                    kwargs["desc"] = (
                        "Calculating pearson correlation with target variable"
                    )
            else:
                self.config.method = "pearsonr"
                kwargs["desc"] = "Calculating pearson correlation with target variable"
            self.func = getattr(scipy.stats, self.config.method)
        return X, kwargs

    def _transform_sklearn(self, X, y):
        cols = DataHandler.get_column_names(X, generate_cols=True)
        corrs = {}
        for col in cols:
            corr, _ = self.func(
                *self._check_data(
                    DataHandler.select_column(X, col), DataHandler.select_column(y, 0)
                )
            )
            corrs[str(col)] = [corr]
        return corrs



from dataclasses import dataclass, field
from typing import List, Type

import numpy as np

from biofit.integration.biosets import get_feature
from biofit.processing import (
    SelectedColumnTypes,
    SelectedFeatureTypes,
    sync_backup_config,
)
from biofit.utils import logging

from ..transformation import Transformer, TransformerConfig

logger = logging.get_logger(__name__)


@dataclass
class LogTransformerConfig(TransformerConfig):
    processor_name: str = field(default="log", init=False, repr=False)
    _transform_process_desc: str = field(
        default="Applying log transformation", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    base = np.e
    shift = 0
    estimator = None


@dataclass
class LogTransformerConfigForOTU(LogTransformerConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)
    base = 2
    shift = 1


class LogTransformer(Transformer):
    output_dtype = "float64"
    config_class = LogTransformerConfig
    config: LogTransformerConfig

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        func = None
        if self.config.base == np.e:
            if self.config.shift == 0:
                func = np.log
            elif self.config.shift == 1:
                func = np.log1p
        elif self.config.base == 10:
            func = np.log10
        elif self.config.base == 2:
            func = np.log2

        if func is None:

            def func(x):
                return np.log(x + self.config.shift) / np.log(self.config.base)

        self.log_transform = func
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "LogTransformer":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedFeatureTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _transform_sklearn(self, X):
        return self.log_transform(X)



import importlib
import importlib.metadata
import importlib.util
import logging
import os
from pathlib import Path

from packaging import version

logger = logging.getLogger(__name__)

ENV_VARS_TRUE_VALUES = {"1", "ON", "YES", "TRUE"}
ENV_VARS_FALSE_VALUES = {"0", "OFF", "NO", "FALSE"}
ENV_VARS_TRUE_AND_AUTO_VALUES = ENV_VARS_TRUE_VALUES.union({"AUTO"})
ENV_VARS_FALSE_AND_AUTO_VALUES = ENV_VARS_FALSE_VALUES.union({"AUTO"})


DEFAULT_XDG_CACHE_HOME = "~/.cache"
XDG_CACHE_HOME = os.getenv("XDG_CACHE_HOME", DEFAULT_XDG_CACHE_HOME)
DEFAULT_BIOFIT_CACHE_HOME = os.path.join(XDG_CACHE_HOME, "biofit")
BIOFIT_CACHE_HOME = os.path.expanduser(
    os.getenv("BIOFIT_HOME", DEFAULT_BIOFIT_CACHE_HOME)
)

DEFAULT_BIOFIT_DATASETS_CACHE = os.path.join(BIOFIT_CACHE_HOME, "datasets")
BIOFIT_DATASETS_CACHE = Path(
    os.getenv("BIOFIT_DATASETS_CACHE", DEFAULT_BIOFIT_DATASETS_CACHE)
)
DEFAULT_BIOFIT_PREPROCESSORS_CACHE = os.path.join(BIOFIT_CACHE_HOME, "processors")
BIOFIT_PROCESSORS_CACHE = Path(
    os.getenv("BIOFIT_PROCESSORS_CACHE", DEFAULT_BIOFIT_PREPROCESSORS_CACHE)
)

DEFAULT_BIOFIT_METRICS_CACHE = os.path.join(BIOFIT_CACHE_HOME, "metrics")
BIOFIT_METRICS_CACHE = Path(
    os.getenv("BIOFIT_METRICS_CACHE", DEFAULT_BIOFIT_METRICS_CACHE)
)

DEFAULT_BIOFIT_MODULES_CACHE = os.path.join(BIOFIT_CACHE_HOME, "modules")
BIOFIT_MODULES_CACHE = Path(
    os.getenv("BIOFIT_MODULES_CACHE", DEFAULT_BIOFIT_MODULES_CACHE)
)

BIOFIT_DYNAMIC_MODULE_NAME = Path(
    os.getenv("BIOFIT_DYNAMIC_MODULE_NAME", "datasets_modules")
)

BIOFIT_CACHE_HOME = os.getenv(
    "BIOFIT_CACHE_HOME",
    os.path.join(os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), "biofit"),
)
DEFAULT_BIOFIT_PATCHES_CACHE = os.path.join(BIOFIT_CACHE_HOME, "patches")
BIOFIT_PATCHES_CACHE = Path(
    os.getenv("BIOFIT_PATCHES_CACHE", DEFAULT_BIOFIT_PATCHES_CACHE)
)

DOWNLOADED_PATCHES_DIR = "downloads"
DEFAULT_DOWNLOADED_PATCHES_PATH = os.path.join(
    BIOFIT_PATCHES_CACHE, DOWNLOADED_PATCHES_DIR
)
DOWNLOADED_PATCHES_PATH = Path(
    os.getenv("BIOFIT_PATCHES_DOWNLOADED_PATCHES_PATH", DEFAULT_DOWNLOADED_PATCHES_PATH)
)

EXTRACTED_PATCHES_DIR = "extracted"
DEFAULT_EXTRACTED_PATCHES_PATH = os.path.join(
    DEFAULT_DOWNLOADED_PATCHES_PATH, EXTRACTED_PATCHES_DIR
)
EXTRACTED_PATCHES_PATH = Path(
    os.getenv("BIOFIT_PATCHES_EXTRACTED_PATCHES_PATH", DEFAULT_EXTRACTED_PATCHES_PATH)
)

PATCHES_FILENAME = "patches.json"
NO_PATCHES_FILENAME = "no_patches.json"
IS_CONDA = os.getenv("CONDA_PREFIX") is not None


PYARROW_AVAILABLE = False
PYARROW_VERSION = "N/A"

PYARROW_AVAILABLE = importlib.util.find_spec("pyarrow") is not None
if PYARROW_AVAILABLE:
    try:
        PYARROW_VERSION = version.parse(importlib.metadata.version("pyarrow"))
        logger.info(f"pyarrow version {PYARROW_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass

RPY2_AVAILABLE = False
RPY2_VERSION = "N/A"

RPY2_AVAILABLE = importlib.util.find_spec("rpy2") is not None
if RPY2_AVAILABLE:
    try:
        RPY2_VERSION = version.parse(importlib.metadata.version("rpy2"))
        logger.info(f"rpy2 version {RPY2_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass

RPY2_ARROW_AVAILABLE = False
RPY2_ARROW_VERSION = "N/A"

RPY2_ARROW_AVAILABLE = importlib.util.find_spec("rpy2_arrow") is not None
if RPY2_ARROW_AVAILABLE:
    try:
        RPY2_ARROW_VERSION = version.parse(importlib.metadata.version("rpy2-arrow"))
        logger.info(f"rpy2 version {RPY2_ARROW_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass

POLARS_VERSION = "N/A"
POLARS_AVAILABLE = False

POLARS_AVAILABLE = importlib.util.find_spec("polars") is not None
if POLARS_AVAILABLE:
    try:
        POLARS_VERSION = version.parse(importlib.metadata.version("polars"))
        logger.info(f"Polars version {POLARS_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass

DASK_VERSION = "N/A"
DASK_AVAILABLE = False

DASK_AVAILABLE = importlib.util.find_spec("dask") is not None
if DASK_AVAILABLE:
    try:
        DASK_VERSION = version.parse(importlib.metadata.version("dask"))
        logger.info(f"Dask version {DASK_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass


USE_TF = os.environ.get("USE_TF", "AUTO").upper()
USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()
USE_JAX = os.environ.get("USE_JAX", "AUTO").upper()


TORCH_VERSION = "N/A"
TORCH_AVAILABLE = False

if USE_TORCH in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TF not in ENV_VARS_TRUE_VALUES:
    TORCH_AVAILABLE = importlib.util.find_spec("torch") is not None
    if TORCH_AVAILABLE:
        try:
            TORCH_VERSION = version.parse(importlib.metadata.version("torch"))
            logger.info(f"PyTorch version {TORCH_VERSION} available.")
        except importlib.metadata.PackageNotFoundError:
            pass
else:
    logger.info("Disabling PyTorch because USE_TF is set")

TF_VERSION = "N/A"
TF_AVAILABLE = False

if USE_TF in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TORCH not in ENV_VARS_TRUE_VALUES:
    TF_AVAILABLE = importlib.util.find_spec("tensorflow") is not None
    if TF_AVAILABLE:
        # For the metadata, we have to look for both tensorflow and tensorflow-cpu
        for package in [
            "tensorflow",
            "tensorflow-cpu",
            "tensorflow-gpu",
            "tf-nightly",
            "tf-nightly-cpu",
            "tf-nightly-gpu",
            "intel-tensorflow",
            "tensorflow-rocm",
            "tensorflow-macos",
        ]:
            try:
                TF_VERSION = version.parse(importlib.metadata.version(package))
            except importlib.metadata.PackageNotFoundError:
                continue
            else:
                break
        else:
            TF_AVAILABLE = False
    if TF_AVAILABLE:
        if TF_VERSION.major < 2:
            logger.info(
                f"TensorFlow found but with version {TF_VERSION}. `datasets` requires version 2 minimum."
            )
            TF_AVAILABLE = False
        else:
            logger.info(f"TensorFlow version {TF_VERSION} available.")
else:
    logger.info("Disabling Tensorflow because USE_TORCH is set")


JAX_VERSION = "N/A"
JAX_AVAILABLE = False

if USE_JAX in ENV_VARS_TRUE_AND_AUTO_VALUES:
    JAX_AVAILABLE = (
        importlib.util.find_spec("jax") is not None
        and importlib.util.find_spec("jaxlib") is not None
    )
    if JAX_AVAILABLE:
        try:
            JAX_VERSION = version.parse(importlib.metadata.version("jax"))
            logger.info(f"JAX version {JAX_VERSION} available.")
        except importlib.metadata.PackageNotFoundError:
            pass
else:
    logger.info("Disabling JAX because USE_JAX is set to False")


USE_BEAM = os.environ.get("USE_BEAM", "AUTO").upper()
BEAM_VERSION = "N/A"
BEAM_AVAILABLE = False
if USE_BEAM in ENV_VARS_TRUE_AND_AUTO_VALUES:
    try:
        BEAM_VERSION = version.parse(importlib.metadata.version("apache_beam"))
        BEAM_AVAILABLE = True
        logger.info(f"Apache Beam version {BEAM_VERSION} available.")
    except importlib.metadata.PackageNotFoundError:
        pass
else:
    logger.info("Disabling Apache Beam because USE_BEAM is set to False")


R_SCRIPTS = Path(__file__).parent / "integration/R/scripts"
BIOFIT_SKIP_R_DEPENDENCIES = (
    os.getenv("BIOFIT_SKIP_R_DEPENDENCIES", "true").upper() not in ENV_VARS_FALSE_VALUES
)
RECORDER_ENABLED = os.getenv("BIOFIT_RECORDER_ENABLED", "true").lower() == "true"

NO_IMBALANCE_ADJUSTMENT = os.getenv("BIOFIT_NO_IMBALANCE_ADJUSTMENT", "false") == "true"

PBAR_REFRESH_TIME_INTERVAL = 0.05



"""
NOTE: The contents of this file have been inlined from the fingerprint and _dill module in the datasets package's source code
https://github.com/huggingface/datasets/blob/c47cc141c9e6e0edafffdcfde55b171612f1de76/src/datasets/fingerprint.py

This module has fixes / adaptations for biofit use cases that make it different from the original
datasets library

The following modifications have been made:
    - Added python source code parsing logic to the `Hasher` class. This is used to hash the contents of a python file
      without comments since we only care about the code that is actually executed.
    - Added the _hash_python_lines function to this file without importing
      huggingface_hub

datasets
~~~~~~~~
Copyright 2023 The HuggingFace Team. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import functools
import hashlib
import importlib
import inspect
import os
import posixpath
import random
import re
from pathlib import Path
import sys
from types import FunctionType, ModuleType
from typing import Any, Dict, List, Union

from biocore.utils.import_util import is_datasets_available
import xxhash
from .version import Version
from ._dill import dumps

import biofit.config as config
import biofit.utils.logging as logging
from biofit.utils.version import __version__

from .file_utils import is_file_name, is_remote_url

_CACHING_ENABLED = True

logger = logging.get_logger(__name__)

fingerprint_rng = random.Random()


def _hash_python_lines(lines: List[str]) -> str:
    filtered_lines = []
    for line in lines:
        line = re.sub(r"#.*", "", line)  # remove comments
        if line:
            filtered_lines.append(line)
    full_str = "\n".join(filtered_lines)

    # Make a hash from all this code
    full_bytes = full_str.encode("utf-8")
    _kwargs = {"usedforsecurity": False} if sys.version_info >= (3, 9) else {}
    sha256 = functools.partial(hashlib.sha256, **_kwargs)
    return sha256(full_bytes).hexdigest()


def generate_random_fingerprint(nbits: int = 64) -> str:
    return f"{fingerprint_rng.getrandbits(nbits):0{nbits//4}x}"


class Hasher:
    """Hasher that accepts python objects as inputs."""

    dispatch: Dict = {}

    def __init__(self):
        self.m = xxhash.xxh64()

    @classmethod
    def hash_bytes(cls, value: Union[bytes, List[bytes]]) -> str:
        value = [value] if isinstance(value, bytes) else value
        m = xxhash.xxh64()
        for x in value:
            m.update(x)
        return m.hexdigest()

    @classmethod
    def hash(cls, value: Any) -> str:
        return cls.hash_bytes(dumps(value))

    @staticmethod
    def _hash_python_lines(self, module: Union[ModuleType, FunctionType, type]) -> str:
        filtered_lines = []
        lines = inspect.getsource(module).splitlines()
        for line in lines:
            line = re.sub(r"#.*", "", line)  # remove comments
            if line:
                filtered_lines.append(line)
        return "\n".join(filtered_lines)

    def update(self, value: Any) -> None:
        header_for_update = f"=={type(value)}=="
        value_for_update = self.hash(value)
        self.m.update(header_for_update.encode("utf8"))
        self.m.update(value_for_update.encode("utf-8"))

    def hexdigest(self) -> str:
        return self.m.hexdigest()


def enable_caching():
    """
    When applying transforms on a dataset, the data are stored in cache files.
    The caching mechanism allows to reload an existing cache file if it's already been computed.

    Reloading a dataset is possible since the cache files are named using the dataset fingerprint, which is updated
    after each transform.

    If disabled, the library will no longer reload cached datasets files when applying transforms to the datasets.
    More precisely, if the caching is disabled:
    - cache files are always recreated
    - cache files are written to a temporary directory that is deleted when session closes
    - cache files are named using a random hash instead of the dataset fingerprint
    - use [`~datasets.Dataset.save_to_disk`] to save a transformed dataset or it will be deleted when session closes
    - caching doesn't affect [`~datasets.load_dataset`]. If you want to regenerate a dataset from scratch you should use
    the `download_mode` parameter in [`~datasets.load_dataset`].
    """
    global _CACHING_ENABLED
    _CACHING_ENABLED = True


def disable_caching():
    """
    When applying transforms on a dataset, the data are stored in cache files.
    The caching mechanism allows to reload an existing cache file if it's already been computed.

    Reloading a dataset is possible since the cache files are named using the dataset fingerprint, which is updated
    after each transform.

    If disabled, the library will no longer reload cached datasets files when applying transforms to the datasets.
    More precisely, if the caching is disabled:
    - cache files are always recreated
    - cache files are written to a temporary directory that is deleted when session closes
    - cache files are named using a random hash instead of the dataset fingerprint
    - use [`~datasets.Dataset.save_to_disk`] to save a transformed dataset or it will be deleted when session closes
    - caching doesn't affect [`~datasets.load_dataset`]. If you want to regenerate a dataset from scratch you should use
    the `download_mode` parameter in [`~datasets.load_dataset`].
    """
    global _CACHING_ENABLED
    _CACHING_ENABLED = False


def is_caching_enabled() -> bool:
    """
    When applying transforms on a dataset, the data are stored in cache files.
    The caching mechanism allows to reload an existing cache file if it's already been computed.

    Reloading a dataset is possible since the cache files are named using the dataset fingerprint, which is updated
    after each transform.

    If disabled, the library will no longer reload cached datasets files when applying transforms to the datasets.
    More precisely, if the caching is disabled:
    - cache files are always recreated
    - cache files are written to a temporary directory that is deleted when session closes
    - cache files are named using a random hash instead of the dataset fingerprint
    """
    global _CACHING_ENABLED
    return bool(_CACHING_ENABLED)


def fingerprint_from_kwargs(fingerprint, kwargs):
    hash = Hasher()
    if fingerprint:
        hash.update(fingerprint)

    for key, value in kwargs.items():
        if isinstance(key, str) and "fingerprint" in key:
            continue
        hash.update(key)
        if isinstance(value, dict):
            hash.update(fingerprint_from_kwargs(fingerprint, value))
        else:
            hash.update(str(value))

    return hash.hexdigest()


def fingerprint_from_data(data):
    from biocore import DataHandler, get_data_format

    if hasattr(data, "_fingerprint"):
        return data._fingerprint
    if hasattr(data, "fingerprint"):
        return data.fingerprint
    hasher = Hasher()
    original_format = get_data_format(data)

    if original_format is not None:
        features = DataHandler.get_column_names(data, generate_cols=True)
        n_samples = DataHandler.get_shape(data)[0]
        first_row = DataHandler.select_row(data, 0)
        hasher.update(original_format)
        hasher.update(features)
        hasher.update(n_samples)
        hasher.update(first_row)
    elif hasattr(data, "__dict__"):
        state = data.__dict__
        for key in sorted(state):
            hasher.update(key)
            try:
                hasher.update(state[key])
            except Exception:
                hasher.update(str(state[key]))
    else:
        raise ValueError("Data object is not hashable")
    return hasher.hexdigest()


def generate_cache_dir(
    X, fingerprint, cache_dir=None, root_dir=config.BIOFIT_DATASETS_CACHE
):
    if cache_dir is None:
        if isinstance(root_dir, Path):
            root_dir = root_dir.as_posix()
        if (
            root_dir == config.BIOFIT_DATASETS_CACHE.as_posix()
            and hasattr(X, "cache_files")
            and X.cache_files
        ):
            cache_dir = os.path.dirname(X.cache_files[0]["filename"])
        else:
            cache_dir = _build_cache_dir(
                X,
                fingerprint,
                cache_dir_root=root_dir,
            )

    if cache_dir:
        Path(cache_dir).mkdir(parents=True, exist_ok=True)
        return Path(cache_dir).resolve().as_posix()
    return None


def get_cache_file_name(cache_dir, fingerprint, cache_file_name=None):
    if cache_file_name:
        if is_file_name(cache_file_name):
            cache_file_name = Path(cache_dir) / Path(cache_file_name).with_suffix(
                ".json"
            )
        else:
            cache_file_name = Path(cache_file_name).with_suffix(".json")
    else:
        cache_file_name = Path(cache_dir) / f"cache-{fingerprint}.json"
    return cache_file_name.resolve().as_posix()


def _relative_data_dir(
    fingerprint, builder_name, dataset_name, version=None, with_hash=True
) -> str:
    """
    Constructs a relative directory path for a dataset based on its properties.

    Args:
        dataset (Dataset): The dataset for which to construct the path.
        with_version (bool, optional): Include version information in the path.
        with_hash (bool, optional): Include hash information in the path.

    Returns:
        str: Relative path for the dataset.
    """
    builder_data_dir = posixpath.join(builder_name, f"{dataset_name}-{fingerprint}")
    if version:
        version = str(version) if isinstance(version, str) else __version__
        builder_data_dir = posixpath.join(builder_data_dir, version)
    if with_hash and is_datasets_available():
        hash = _hash_python_lines(
            inspect.getsource(
                getattr(importlib.import_module("datasets.table"), "InMemoryTable")
            )
        )
        builder_data_dir = posixpath.join(builder_data_dir, hash)
    return builder_data_dir


def _build_cache_dir(
    obj,
    fingerprint: str,
    cache_dir_root: str = config.BIOFIT_DATASETS_CACHE,
) -> str:
    """
    Builds the cache directory path for storing processed dataset versions.

    Args:
        dataset (Union[Dataset, IterableDataset]): The dataset to cache.
        cache_dir_root (str, optional): Root directory for caching datasets.

    Returns:
        str: The path to the dataset's cache directory.
    """
    if (
        hasattr(obj, "version")
        and hasattr(obj, "config_name")
        and hasattr(obj, "builder_name")
    ):
        version = str(obj.version) if isinstance(obj.version, str) else __version__
        dataset_name = obj.config_name or "default"
        builder_name = obj.builder_name or "in_memory"
    elif (
        hasattr(obj, "config")
        and hasattr(obj.config, "version")
        and hasattr(obj.config, "processor_name")
        and hasattr(obj.config, "processor_type")
    ):
        version = (
            str(obj.config.version)
            if obj.config.version and not str(obj.config.version) == "0.0.0"
            else __version__
        )
        dataset_name = obj.config.processor_name or "default"
        builder_name = obj.config.processor_type or "in"

    else:
        version = __version__
        dataset_name = "default"
        builder_name = "in_memory"

    builder_data_dir = posixpath.join(
        cache_dir_root,
        _relative_data_dir(
            fingerprint=fingerprint,
            builder_name=builder_name,
            dataset_name=dataset_name,
        ),
    )
    version_data_dir = posixpath.join(
        cache_dir_root,
        _relative_data_dir(
            fingerprint=fingerprint,
            builder_name=builder_name,
            dataset_name=dataset_name,
            version=version,
        ),
    )

    def _other_versions_on_disk():
        """Returns previous versions on disk."""
        if not os.path.exists(builder_data_dir):
            return []

        version_dirnames = []
        for dir_name in os.listdir(builder_data_dir):
            try:
                version_dirnames.append((Version(dir_name), dir_name))
            except ValueError:  # Invalid version (ex: incomplete data dir)
                pass
        version_dirnames.sort(reverse=True)
        return version_dirnames
        # Check and warn if other versions exist

    if not is_remote_url(builder_data_dir):
        version_dirs = _other_versions_on_disk()
        if version_dirs:
            other_version = version_dirs[0][0]
            if other_version != version:
                warn_msg = (
                    f"Found a different version {str(other_version)} of dataset {dataset_name} in "
                    f"cache_dir {cache_dir_root}. Using currently defined version "
                    f"{str(version)}."
                )
                logger.warning(warn_msg)
    if not os.path.exists(version_data_dir):
        os.makedirs(version_data_dir, exist_ok=True)

    return version_data_dir


def update_fingerprint(fingerprint, value, key=None):
    if key is None and value is None:
        return fingerprint
    hasher = Hasher()
    if fingerprint:
        hasher.update(fingerprint)
    if key:
        hasher.update(key)
    try:
        hasher.update(value)
    except Exception:
        return generate_random_fingerprint()
    else:
        return hasher.hexdigest()



from dataclasses import dataclass, field

from biofit.integration.biosets import get_feature
from biofit.processing import BaseProcessor, ProcessorConfig, SelectedFeatureTypes


@dataclass
class FeatureExtractorConfig(ProcessorConfig):
    """Base class for feature extraction processor configurations."""

    processor_type: str = field(default="feature_extraction", init=False, repr=False)
    _fit_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )


class FeatureExtractor(BaseProcessor):
    """Base class for feature extraction processors."""



import argparse
import sys

import biofit.cli.install_r as install_r
from biofit.utils import logging

logger = logging.get_logger(__name__)


def create_parser():
    parser = argparse.ArgumentParser(
        description="BIOFIT: General-purpose Omics Machine Learning framework"
    )

    # add a quiet option
    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Suppress all logging output except for errors",
    )
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    return parser, subparsers


def main():
    parser, subparsers = create_parser()
    subparsers = install_r.add_to_parser(subparsers)
    args = parser.parse_args()

    if args.quiet:
        logging.set_verbosity(logging.ERROR)
    else:
        logging.set_verbosity(logging.INFO)

    try:
        if args.subcommand == "install":
            install_r.run(args)
        else:
            parser.print_help()
            sys.exit(1)
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type

import pyarrow as pa
from biocore.utils.import_util import is_datasets_available

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils.types import Unset

from ..plot_feature_selection import (
    FeatureSelectorPlotter,
    FeatureSelectorPlotterConfig,
)

if TYPE_CHECKING:
    pass


@dataclass
class MinPrevalencePlotterConfig(FeatureSelectorPlotterConfig):
    plot_process_desc: str = field(
        default="Plotting presence feature selection", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _compare: bool = field(default=True, init=False, repr=False)
    _add_labels: bool = field(default=False, init=False, repr=False)
    processor_name: str = field(
        default="min_prevalence_feature_selector", init=False, repr=False
    )

    sample_xlab: str = "Sum of Counts"
    sample_main: str = "Sample Distribution"
    feature_xlab: str = "Sum of Counts"
    feature_main: str = "Feature Distribution"
    legend_position: str = "top"
    legend_title: str = "Max Missing Feature Selection"
    before_name: str = "Before"
    after_name: str = "After"
    xlog: str = None
    ylog: str = None
    ncol: int = 2
    include_non_zero_sum: bool = False
    non_zero_samp_xlab: str = None
    non_zero_samp_main: str = None
    non_zero_feat_xlab: str = None
    non_zero_feat_main: str = None


@dataclass
class MinPrevalencePlotterConfigForMetagenomics(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="metagenomics", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )

    feature_main = "Taxa Total Abundance Distribution"
    non_zero_samp_xlab: str = "Species Richness"
    non_zero_samp_main: str = "Richness Distribution"
    non_zero_feat_xlab: str = "Species Prevalence"
    non_zero_feat_main: str = "Prevalence Across Samples"
    xlog = "log2_1p"
    include_non_zero_sum: bool = True


@dataclass
class MinPrevalencePlotterConfigForOTU(MinPrevalencePlotterConfigForMetagenomics):
    dataset_name: str = field(default="otu", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )
    feature_main: str = "OTU Total Abundance Distribution"
    include_non_zero_sum: bool = True


@dataclass
class MinPrevalencePlotterConfigForASV(MinPrevalencePlotterConfigForMetagenomics):
    dataset_name: str = field(default="asv", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance"), get_feature("Abundance")],
        init=False,
        repr=False,
    )
    include_non_zero_sum: bool = True


@dataclass
class MinPrevalencePlotterConfigForGenomics(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="genomics", init=False, repr=False)
    _input_feature_types: List[Type] = field(
        default_factory=lambda: [
            (get_feature("ReadCount"), get_feature("GenomicVariant")),
            (get_feature("ReadCount"), get_feature("GenomicVariant")),
        ],
        init=False,
        repr=False,
    )


@dataclass
class MinPrevalencePlotterConfigForSNP(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="snp", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("GenomicVariant"),
            get_feature("GenomicVariant"),
        ],
        init=False,
        repr=False,
    )


@dataclass
class MinPrevalencePlotterConfigForReadCount(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="read_count", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ReadCount"), get_feature("ReadCount")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ReadCount"), get_feature("ReadCount")],
        init=False,
        repr=False,
    )


@dataclass
class MinPrevalencePlotterConfigForProteomics(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="proteomics", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Expression"), get_feature("Expression")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Expression"), get_feature("Expression")],
        init=False,
        repr=False,
    )


@dataclass
class MinPrevalencePlotterConfigForMaldi(MinPrevalencePlotterConfig):
    dataset_name: str = field(default="maldi", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("PeakIntensity"),
            get_feature("PeakIntensity"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("PeakIntensity"),
            get_feature("PeakIntensity"),
        ],
        init=False,
        repr=False,
    )


class MinPrevalenceFeatureSelectorPlotter(FeatureSelectorPlotter):
    config_class = MinPrevalencePlotterConfig
    config: MinPrevalencePlotterConfig

    def __init__(
        self,
        sample_xlab: str = "Sum of Counts",
        sample_main: str = "Sample Distribution",
        feature_xlab: str = "Sum of Counts",
        feature_main: str = "Feature Distribution",
        legend_position: str = "top",
        legend_title: str = "Max Missing Feature Selection",
        before_name: str = "Before",
        after_name: str = "After",
        xlog: str = None,
        ylog: str = None,
        ncol: int = 2,
        include_non_zero_sum: bool = False,
        non_zero_samp_xlab: str = None,
        non_zero_samp_main: str = None,
        non_zero_feat_xlab: str = None,
        non_zero_feat_main: str = None,
        config: MinPrevalencePlotterConfig = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            sample_xlab=sample_xlab,
            sample_main=sample_main,
            feature_xlab=feature_xlab,
            feature_main=feature_main,
            legend_position=legend_position,
            legend_title=legend_title,
            before_name=before_name,
            after_name=after_name,
            xlog=xlog,
            ylog=ylog,
            ncol=ncol,
            include_non_zero_sum=include_non_zero_sum,
            non_zero_samp_xlab=non_zero_samp_xlab,
            non_zero_samp_main=non_zero_samp_main,
            non_zero_feat_xlab=non_zero_feat_xlab,
            non_zero_feat_main=non_zero_feat_main,
            **kwargs,
        )

    def plot_pandas(self, x1, x2):
        row_sums = [pa.array(x1.sum(axis=1)), pa.array(x2.sum(axis=1))]
        col_sums = [pa.array(x1.sum(axis=0)), pa.array(x2.sum(axis=0))]
        input = [row_sums, col_sums]
        mains = [self.config.sample_main, self.config.feature_main]
        xlabs = [self.config.sample_xlab, self.config.feature_xlab]
        if self.config.include_non_zero_sum:
            non_zero_sample_sums = [
                pa.array(x1[x1 > 0].count(axis=1)),
                pa.array(x2[x2 > 0].count(axis=1)),
            ]
            non_zero_feature_sums = [
                pa.array(x1[x1 > 0].count(axis=0)),
                pa.array(x2[x2 > 0].count(axis=0)),
            ]
            mains.extend(
                [
                    self.config.non_zero_samp_main,
                    self.config.non_zero_feat_main,
                ]
            )
            xlabs.extend(
                [
                    self.config.non_zero_samp_xlab,
                    self.config.non_zero_feat_xlab,
                ]
            )
            input.extend([non_zero_sample_sums, non_zero_feature_sums])

        self.plotter(
            list_of_sums=input,
            path=self.config.path,
            xlabs=xlabs,
            mains=mains,
            ncol=self.config.ncol,
            legend_position=self.config.legend_position,
            legend_title=self.config.legend_title,
            before_name=self.config.before_name,
            after_name=self.config.after_name,
            xlog=self.config.xlog,
            ylog=self.config.ylog,
        )

    def plot(
        self,
        x1,
        x2=None,
        input_columns1: SelectedColumnTypes = None,
        input_columns2: SelectedColumnTypes = None,
        sample_xlab: str = Unset('"Sum of Counts"'),
        sample_main: str = Unset('"Sample Distribution"'),
        feature_xlab: str = Unset('"Sum of Counts"'),
        feature_main: str = Unset('"Feature Distribution"'),
        legend_position: str = Unset('"top"'),
        legend_title: str = Unset('"Max Missing Feature Selection"'),
        before_name: str = Unset('"Before"'),
        after_name: str = Unset('"After"'),
        xlog: str = Unset("None"),
        ylog: str = Unset("None"),
        ncol: int = Unset("2"),
        include_non_zero_sum: bool = Unset("False"),
        non_zero_samp_xlab: str = Unset("None"),
        non_zero_samp_main: str = Unset("None"),
        non_zero_feat_xlab: str = Unset("None"),
        non_zero_feat_main: str = Unset("None"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        if x2 is None:
            if is_datasets_available():
                from biosets import Bioset
                from datasets import Dataset as HfDataset

                if isinstance(x1, (Bioset, HfDataset)):
                    _, _, prev_replays = self._get_replays(x1)
                    if prev_replays:
                        x2 = Bioset.from_replays(prev_replays)
                    return self._plot(x2, x1)
            else:
                raise ValueError(
                    "Must provide the before and after feature selection datasets."
                )

        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns1, input_columns2
        )
        return self._plot(
            x1,
            x2,
            sample_xlab=sample_xlab,
            sample_main=sample_main,
            feature_xlab=feature_xlab,
            feature_main=feature_main,
            legend_position=legend_position,
            legend_title=legend_title,
            before_name=before_name,
            after_name=after_name,
            xlog=xlog,
            ylog=ylog,
            ncol=ncol,
            include_non_zero_sum=include_non_zero_sum,
            non_zero_samp_xlab=non_zero_samp_xlab,
            non_zero_samp_main=non_zero_samp_main,
            non_zero_feat_xlab=non_zero_feat_xlab,
            non_zero_feat_main=non_zero_feat_main,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )



import json
import warnings
from dataclasses import dataclass, field

import numpy as np
import pandas as pd
import pyarrow as pa
from biocore import DataHandler
from biocore.utils.import_util import requires_backends

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, SelectedFeatureTypes
from biofit.utils.types import Unset
from biofit.visualization.plotting import BasePlotter, PlotterConfig
from biofit.visualization.plotting_utils import get_distinct_colors


@dataclass
class DimensionReducerPlotterConfig(PlotterConfig):
    processor_type: str = field(default="feature_extractor", init=False, repr=False)
    _fit_input_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [
            None,
            get_feature("TARGET_FEATURE_TYPES"),
            get_feature("METADATA_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )

    title: str = "Dimension Reduction Plot"
    colormap: str = "nipy_spectral"
    n_components: int = 3
    label_column: str = None
    group_column: str = None


class DimensionReductionPlotter(BasePlotter):
    """Base class for feature extraction processors."""

    _config_class = DimensionReducerPlotterConfig
    config: DimensionReducerPlotterConfig

    def __init__(
        self,
        label_column: str = None,
        group_column: str = None,
        title: str = Unset('"Dimension Reduction Plot"'),
        colormap: str = Unset('"nipy_spectral"'),
        n_components: int = Unset("3"),
        config: DimensionReducerPlotterConfig = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            n_components=n_components,
            label_column=label_column,
            group_column=group_column,
            title=title,
            colormap=colormap,
            **kwargs,
        )

    def plot(
        self,
        X,
        labels=None,
        group=None,
        input_columns: SelectedColumnTypes = None,
        label_column: SelectedColumnTypes = None,
        group_column: SelectedColumnTypes = None,
        title: str = Unset('"Dimension Reduction Plot"'),
        colormap: str = Unset('"nipy_spectral"'),
        n_components: int = Unset("3"),
        dimension_reducer=None,
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show=True,
    ):
        """Plot the PCoA plot.

        Args:
            X (array_like):
                The input data. Can be a Dataset, polars/pandas DataFrame, numpy array, or Arrow table.
            labels (array_like, *optional*):
                The labels for the data. Dictates the color of the points in the plot.
                Must be an array of values with the same length as the number of rows in X.
                If not provided, the points will be colored by group.
            group (array_like, *optional*):
                The group for the data. Dictates the shape of the points in the plot.
                Must be an array of values with the same length as the number of rows in X.
            pcoa (array_like, *optional*):
                The fitted PCoAFeatureExtractor object. Used to extract the eigvals for the explained variance ratio.
                If not provided, the explained variance ratio will not be displayed.
            **kwargs:
                Additional keyword arguments to pass to the plot:
                    n_components (int, 2):
                        The number of components to plot. Defaults to 2.
                    label_name (str, *optional*):
                        The name of the labels. Used for the legend and retrieving the values from X if labels is not provided.
                    group_name (str, *optional*):
                        The name of the group. Used for the legend and retrieving the values from X if group is not provided.

        """
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, label_column, group_column
        )
        return self._plot(
            X,
            labels,
            group,
            n_components=n_components,
            dimension_reducer=dimension_reducer,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_dataset(self, X, labels=None, group=None, dimension_reducer=None):
        from biosets import decode

        if labels is not None:
            labels = decode(labels)

        return self.plot_arrow(
            DataHandler.to_arrow(X),
            DataHandler.to_arrow(labels) if labels is not None else None,
            DataHandler.to_arrow(group) if group is not None else None,
            dimension_reducer,
        )

    def plot_arrow(self, X, labels=None, group=None, dimension_reducer=None):
        def get_int2str_converter(x) -> np.ndarray:
            if not isinstance(x, pa.Table):
                return lambda x: x

            def get_features(x: pa.Table):
                metadata = x.schema.metadata
                if metadata:
                    if b"huggingface" in metadata:
                        metadata = json.loads(metadata[b"huggingface"].decode("utf-8"))

                    if "info" in metadata and "features" in metadata["info"]:
                        return metadata["info"]["features"]

                return {}

            metadata = get_features(x)
            lab_col = x.column_names[0]
            if metadata and lab_col in metadata:
                label_metadata = metadata[lab_col]
                if label_metadata["_type"] == "ClassLabel":
                    label_metadata.pop("_type")
                    cls_label = get_feature("ClassLabel")(**label_metadata)
                    return cls_label.int2str
            return lambda x: x

        def pairplot(
            tbl: pd.DataFrame,
            n_components,
            ex_var_ratio=None,
            label_name=None,
            group_name=None,
            marker_dict=None,
            color_dict=None,
        ):
            requires_backends(pairplot, "seaborn")
            requires_backends(pairplot, "matplotlib")
            import matplotlib.pyplot as plt
            import seaborn as sns
            from matplotlib.lines import Line2D

            features = tbl.columns[:n_components]
            fig, axes = plt.subplots(n_components, n_components, figsize=(15, 15))

            for i, f1 in enumerate(features):
                for j, f2 in enumerate(features):
                    ax = axes[i, j]
                    if i != j:
                        if label_name is not None and group_name is not None:
                            for (lab, grp), df_group in tbl.groupby(
                                [
                                    label_name,
                                    group_name,
                                ]
                            ):
                                sns.scatterplot(
                                    x=f2,
                                    y=f1,
                                    data=df_group,
                                    ax=ax,
                                    color=color_dict[lab],
                                    marker=marker_dict[grp],
                                )
                        elif label_name is not None:
                            for lab, df_lab in tbl.groupby(label_name):
                                sns.scatterplot(
                                    x=f2,
                                    y=f1,
                                    data=df_lab,
                                    ax=ax,
                                    color=color_dict[lab],
                                )
                        elif group_name is not None:
                            for grp, df_group in tbl.groupby(group_name):
                                sns.scatterplot(
                                    x=f2,
                                    y=f1,
                                    data=df_group,
                                    ax=ax,
                                    marker=marker_dict[grp],
                                )
                        else:
                            sns.scatterplot(x=f2, y=f1, data=tbl, ax=ax)

                    else:
                        if label_name is not None:
                            for lab, df_lab in tbl.groupby(label_name):
                                sns.kdeplot(
                                    df_lab[f1],
                                    ax=ax,
                                    color=color_dict[lab],
                                    fill=True,
                                )
                        else:
                            sns.kdeplot(tbl[f1], ax=ax, fill=True)
                        if ex_var_ratio is not None:
                            plt.text(
                                0.1,
                                0.9,
                                f"Dim{i + 1}: {float(ex_var_ratio[i]) * 100:.2f}%",
                                transform=ax.transAxes,
                            )
                    # Set only left and bottom borders
                    ax.spines["top"].set_visible(False)
                    ax.spines["right"].set_visible(False)

                    # Enable ticks on left and bottom only
                    ax.xaxis.set_tick_params(which="both", bottom=True)
                    ax.yaxis.set_tick_params(which="both", left=True)

                    # Set axis labels
                    if i == n_components - 1:
                        ax.set_xlabel(f2)
                    else:
                        ax.set_xlabel("")
                        ax.set_xticklabels([])

                    if j == 0:
                        ax.set_ylabel(f1)
                    else:
                        ax.set_ylabel("")
                        ax.set_yticklabels([])

            if label_name:
                label_lines = [
                    Line2D(
                        [0],
                        [0],
                        color=color,
                        marker="o",
                        linestyle="",
                        markersize=10,
                    )
                    for color in color_dict.values()
                ]
                u_labels = list(color_dict.keys())
                n_labels = len(u_labels)
                label_height = 0.5 + 0.216 / 15 * (n_labels + 1) / 2 + 0.01
                label_legend = fig.legend(
                    label_lines,
                    u_labels,
                    title=label_name,
                    loc="center left",
                    bbox_to_anchor=(0.90, label_height),
                )

            if group_name:
                group_lines = [
                    Line2D(
                        [0],
                        [0],
                        color="gray",
                        marker=marker,
                        linestyle="",
                        markersize=10,
                    )
                    for marker in marker_dict.values()
                ]
                u_group = list(marker_dict.keys())
                n_group = len(u_group)

                group_height = 0.5 - 0.216 / 15 * (n_group + 1) / 2 - 0.01
                fig.legend(
                    group_lines,
                    list(marker_dict.keys()),
                    title=group_name,
                    loc="center left",
                    bbox_to_anchor=(0.90, group_height),
                )
                if label_name:
                    fig.add_artist(label_legend)

            if self.config.title:
                fig.suptitle(self.config.title, fontsize=16)

            return fig

        warnings.filterwarnings("ignore")
        n_components = self.config.n_components
        tbl = DataHandler.to_format(X, "pandas").iloc[:, :n_components]
        labs = None
        label_name = self.config.label_column or "labels"
        u_labels = None
        grp = None
        grp_name = self.config.group_column or None
        u_group = None
        if labels is not None:
            labs = DataHandler.to_format(labels, "pandas")
            if isinstance(labs, pd.DataFrame):
                label_name = labs.columns[0] if label_name is None else label_name
                labs = labs.iloc[:, 0]
            elif isinstance(labs, pd.Series):
                label_name = labs.name if label_name is None else label_name
            labs = labs.fillna("None")
            u_labels = np.unique(labs)
        if group is not None:
            grp = DataHandler.to_format(group, "pandas")
            if isinstance(grp, pd.DataFrame):
                grp_name = grp.columns[0] if grp_name is None else grp_name
                grp = grp.iloc[:, 0]
            elif isinstance(grp, pd.Series):
                grp_name = grp.name if grp_name is None else grp_name
            u_group = np.unique(grp)

        if labs is not None:
            converter = get_int2str_converter(labels)
            dtype = DataHandler.get_dtypes(labels)
            if "int" in next(iter(dtype.values())):
                encodings = labs.to_numpy()
                encodings_dims = encodings.shape
                if len(encodings_dims) == 1:
                    labs = converter(encodings)
                elif encodings_dims[1] == 1:
                    labs = converter(encodings[:, 0])

        if labs is not None:
            u_labels = np.unique(labs)

        ex_var_ratio = None
        if dimension_reducer is not None:
            if hasattr(dimension_reducer.config, "eigvals"):
                eigvals = dimension_reducer.config.eigvals
                ex_var_ratio = eigvals / eigvals.sum()
            elif hasattr(dimension_reducer.config, "estimator"):
                ex_var_ratio = (
                    dimension_reducer.config.estimator.explained_variance_ratio_
                )

        tbl.columns = [f"Dim{i + 1}" for i in range(len(tbl.columns))]

        marker_styles = ["o", "v", "^", "<", ">", "s", "p", "*", "H", "X"]
        if u_group is not None:
            marker_dict = {
                label: marker_styles[i % len(marker_styles)]
                for i, label in enumerate(u_group)
            }
        else:
            marker_dict = None

        if labs is not None and grp is not None:
            tbl = tbl.assign(**{label_name: labs, grp_name: grp})
        elif labs is not None:
            tbl = tbl.assign(**{label_name: labs})
        elif grp is not None:
            tbl = tbl.assign(**{grp_name: grp})

        if labs is not None:
            color_dict = {
                label: color
                for label, color in zip(
                    u_labels, get_distinct_colors(len(u_labels), self.config.colormap)
                )
            }
        elif grp is not None:
            color_dict = {
                label: color
                for label, color in zip(
                    u_group, get_distinct_colors(len(u_group), self.config.colormap)
                )
            }
        else:
            color_dict = None

        fig = pairplot(
            tbl,
            n_components,
            ex_var_ratio,
            label_name,
            grp_name,
            marker_dict,
            color_dict,
        )

        warnings.filterwarnings("default")

        # save the figure to the path
        if self.config.path:
            fig.savefig(self.config.path)
        return fig



import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional, Type, Union

import numpy as np
import pandas as pd
from biocore.utils.import_util import is_polars_available

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.stat import RowMissingnessStat, RowMissingnessStatConfig
from biofit.utils import Unset, logging

from ..filtering import SampleFilter, SampleFilterConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class MinPrevalenceRowSampleFilterConfig(SampleFilterConfig):
    # process description
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    processor_name: str = field(
        default="min_prevalence_sample_filter", init=False, repr=False
    )

    # default values
    min_prevalence: float = "auto"
    depth: int = None


@dataclass
class MinPrevalenceRowSampleFilterConfigForOTU(MinPrevalenceRowSampleFilterConfig):
    # dataset description
    dataset_name: str = field(default="otu", init=False, repr=False)

    # override default values
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    depth: int = 0
    min_prevalence: float = "auto"

    def __post_init__(self):
        if self.depth is None:
            self._transform_process_desc = (
                f"Removing samples with <{self.min_prevalence * 100:.0f}% present OTUs"
            )
        elif isinstance(self.min_prevalence, float) and self.min_prevalence < 1:
            self._transform_process_desc = f"Removing samples with <{self.min_prevalence * 100:.0f}% OTUs over {self.depth} counts"
        elif isinstance(self.min_prevalence, (int, float)) and self.min_prevalence >= 1:
            self._transform_process_desc = f"Removing samples with <{self.min_prevalence} OTUs over {self.depth} counts"


@dataclass
class MinPrevalenceRowSampleFilterConfigForSNP(MinPrevalenceRowSampleFilterConfig):
    # dataset description
    dataset_name: str = field(default="snp", init=False, repr=False)

    # override default values
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    depth: int = 0
    min_prevalence: float = 0.012


@dataclass
class MinPrevalenceRowSampleFilterConfigForMaldi(MinPrevalenceRowSampleFilterConfig):
    # dataset description
    dataset_name: str = field(default="maldi", init=False, repr=False)

    # override default values
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("PeakIntensity")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("PeakIntensity")], init=False, repr=False
    )
    depth: int = 0
    min_prevalence: float = 0.4


class MinPrevalenceSampleFilter(SampleFilter):
    # main config class
    config_class = MinPrevalenceRowSampleFilterConfig
    config: MinPrevalenceRowSampleFilterConfig

    IQRs = None

    def __init__(
        self,
        config: Optional[MinPrevalenceRowSampleFilterConfig] = None,
        *,
        min_prevalence: Union[str, float] = "auto",
        depth: Union[int, Unset] = None,
        **kwargs,
    ):
        """SampleFilter samples based on the minimum prevalence of features.

        Args:
            config (MinPrevalenceRowSampleFilterConfig, optional):
                The configuration for the filter. Defaults to None. If a configuration is provided, the
                below arguments are ignored. To override the configuration, use the set_params method.
            min_prevalence (Union[str, float], optional):
                The minimum prevalence of features to keep a sample.
                The threshold to which we determine the minimum prevalence percentage or count of present values to maintain a sample or feature.
                Any value passed >= 1 will be the minimum count while any value < 1 will be a percentage of the total values.
                If "auto", the minimum prevalence is calculated as the first quartile minus 1.5 times the interquartile range. Defaults to "auto".
            depth (Union[int, Unset], optional):
                The minimum value that we consider something to be present. Defaults to None.
            **kwargs:
                Additional keyword arguments to pass to the filter. See the ProcessorConfig class for more details.
        """
        super().__init__(
            config=config, min_prevalence=min_prevalence, depth=depth, **kwargs
        )
        row_missingness_config = RowMissingnessStatConfig.from_config(self.config)
        self.missingness = RowMissingnessStat(row_missingness_config)

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        row_missingness_config = RowMissingnessStatConfig.from_config(self.config)
        self.missingness = RowMissingnessStat(row_missingness_config)
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "MinPrevalenceSampleFilter":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _process_fit_input(self, input, **kwargs):
        if self.config.min_prevalence != "auto":
            kwargs["fn_kwargs"]["fn"] = None
        return super()._process_fit_input(input, **kwargs)

    def _fit_polars(self, X: "pl.DataFrame"):
        row_missingness = self.missingness._transform_polars(X)
        row_missingness = row_missingness.get_column(row_missingness.columns[0])
        row_present = X.shape[1] - row_missingness
        iqr = [row_present.quantile(0.25), row_present.quantile(0.75)]
        self.config.min_prevalence = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_polars(self, X: "pl.DataFrame"):
        row_missingness = self.missingness._transform_polars(X)
        row_missingness = row_missingness.get_column(row_missingness.columns[0])
        row_present = X.shape[1] - row_missingness
        iqr = [row_present.quantile(0.25), row_present.quantile(0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _fit_pandas(self, X: "pd.DataFrame"):
        row_missingness = self.missingness._transform_pandas(X).iloc[:, 0]
        row_present = X.shape[1] - row_missingness
        iqr = [row_present.quantile(0.25), row_present.quantile(0.75)]
        self.config.min_prevalence = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_pandas(self, X: "pd.DataFrame"):
        row_missingness = self.missingness._transform_pandas(X).iloc[:, 0]
        row_present = X.shape[1] - row_missingness
        iqr = [row_present.quantile(0.25), row_present.quantile(0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _fit_numpy(self, X: np.ndarray):
        row_missingness = self.missingness._transform_numpy(X)[:, 0]
        row_present = X.shape[1] - row_missingness
        iqr = [np.quantile(row_present, 0.25), np.quantile(row_present, 0.75)]
        self.config.min_prevalence = iqr[0] - 1.5 * (iqr[1] - iqr[0])
        return self

    def _partial_fit_numpy(self, X: np.ndarray):
        row_missingness = self.missingness._transform_numpy(X)[:, 0]
        row_present = X.shape[1] - row_missingness
        iqr = [np.quantile(row_present, 0.25), np.quantile(row_present, 0.75)]
        if self.IQRs is None:
            self.IQRs = [[iqr[0]], [iqr[1]]]
        else:
            self.IQRs[0].append(iqr[0])
            self.IQRs[1].append(iqr[1])
        return self

    def _pool_fit_any(self, partial_results: List["MinPrevalenceSampleFilter"]):
        IQRs = [[], []]
        for result in partial_results:
            IQRs[0].extend(result.IQRs[0])
            IQRs[1].extend(result.IQRs[1])
        IQRs[0] = np.mean(IQRs[0])
        IQRs[1] = np.mean(IQRs[1])
        self.config.min_prevalence = IQRs[0] - 1.5 * (IQRs[1] - IQRs[0])
        return self

    def _process_fit_output(self, input, out):
        logger.info(f"Minimum prevalence set to {out.config.min_prevalence * 100:.0f}%")
        return super()._process_fit_output(input, out)

    def _transform_polars(self, X: "pl.DataFrame"):
        if is_polars_available() and "polars" in sys.modules:
            import polars as pl
        total_present: pl.DataFrame = self.missingness._transform_polars(
            X
        ).with_columns((X.shape[1] - pl.col("*")).alias("sum"))

        if self.config.min_prevalence < 1:
            total_present = total_present.with_columns(pl.col("sum") / X.shape[1])
        tests = total_present.with_columns(pl.col("sum") > self.config.min_prevalence)
        if len(tests) == 1:
            return tests[0]
        return tests.to_numpy()[:, 0].tolist()

    def _transform_pandas(self, X: pd.DataFrame):
        total_present = X.shape[1] - self.missingness._transform_pandas(X)
        if self.config.min_prevalence < 1:
            total_present = total_present / X.shape[1]
        tests = total_present > self.config.min_prevalence
        if len(tests) == 1:
            return tests[0]
        return tests.values[:, 0].tolist()

    def _transform_numpy(self, X: np.ndarray):
        total_present = X.shape[1] - self.missingness._transform_numpy(X)
        if self.config.min_prevalence < 1:
            total_present = total_present / X.shape[1]
        tests = total_present > self.config.min_prevalence
        if len(tests) == 1:
            return tests[0]
        return tests[:, 0].tolist()



from functools import partial

from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    log_loss,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
)

from .metrics import (
    calculate_metrics,  # noqa: F401
    confusion_matrix,  # noqa: F401
    log_loss_weighted,
    specificity,
)

_OPTUNA_METRICS = {
    "binary_classification": "f1",
    "multiclass_classification": "f1_weighted",
    "regression": "rmse",
    "multi_regression": "rmse",
}

_METRICS = {
    "binary_classification": {
        "logloss": (log_loss, "minimize"),
        "logloss_weighted": (
            partial(log_loss_weighted, class_weights="balanced"),
            "minimize",
        ),
        "auc": (roc_auc_score, "maximize"),
        "f1": (f1_score, "maximize"),
        "accuracy": (accuracy_score, "maximize"),
        "balanced_accuracy": (balanced_accuracy_score, "maximize"),
        "precision": (precision_score, "maximize"),
        "recall": (recall_score, "maximize"),
        "specificity": (specificity, "maximize"),
    },
    "multiclass_classification": {
        "mlogloss": (
            lambda y_true, y_pred, labels: log_loss(y_true, y_pred, labels=labels),
            "minimize",
        ),
        "mlogloss_weighted": (
            lambda y_true, y_pred, labels: log_loss_weighted(
                y_true, y_pred, labels=labels, class_weights="balanced"
            ),
            "minimize",
        ),
        "accuracy": (
            lambda y_true, y_pred, labels: accuracy_score(y_true, y_pred),
            "maximize",
        ),
        "balanced_accuracy": (
            lambda y_true, y_pred, labels: balanced_accuracy_score(y_true, y_pred),
            "maximize",
        ),
        "f1_macro": (
            lambda y_true, y_pred, labels: f1_score(
                y_true, y_pred, average="macro", labels=labels
            ),
            "maximize",
        ),
        "f1_micro": (
            lambda y_true, y_pred, labels: f1_score(
                y_true, y_pred, average="micro", labels=labels
            ),
            "maximize",
        ),
        "f1_weighted": (
            lambda y_true, y_pred, labels: f1_score(
                y_true, y_pred, average="weighted", labels=labels
            ),
            "maximize",
        ),
        "precision_macro": (
            lambda y_true, y_pred, labels: precision_score(
                y_true, y_pred, average="macro", labels=labels
            ),
            "maximize",
        ),
        "precision_micro": (
            lambda y_true, y_pred, labels: precision_score(
                y_true, y_pred, average="micro", labels=labels
            ),
            "maximize",
        ),
        "precision_weighted": (
            lambda y_true, y_pred, labels: precision_score(
                y_true, y_pred, average="weighted", labels=labels
            ),
            "maximize",
        ),
        "recall_macro": (
            lambda y_true, y_pred, labels: recall_score(
                y_true, y_pred, average="macro", labels=labels
            ),
            "maximize",
        ),
        "recall_micro": (
            lambda y_true, y_pred, labels: recall_score(
                y_true, y_pred, average="micro", labels=labels
            ),
            "maximize",
        ),
        "recall_weighted": (
            lambda y_true, y_pred, labels: recall_score(
                y_true, y_pred, average="weighted", labels=labels
            ),
            "maximize",
        ),
        "specificity_macro": (
            lambda y_true, y_pred, labels: specificity(
                y_true, y_pred, average="macro", labels=labels
            ),
            "maximize",
        ),
        # "specificity_micro": (
        #     lambda y_true, y_pred, labels: specificity(
        #         y_true, y_pred, average="micro", labels=labels
        #     ),
        #     "maximize",
        # ),
        "specificity_weighted": (
            lambda y_true, y_pred, labels: specificity(
                y_true, y_pred, average="weighted", labels=labels
            ),
            "maximize",
        ),
    },
    "regression": {
        "rmse": (partial(mean_squared_error, squared=False), "minimize"),
        "rmsle": (partial(mean_squared_log_error, squared=False), "minimize"),
        "r2": (r2_score, "maximize"),
        "mse": (mean_squared_error, "minimize"),
        "mae": (mean_absolute_error, "minimize"),
    },
    "multi_regression": {
        "rmse": (partial(mean_squared_error, squared=False), "minimize"),
        "rmsle": (partial(mean_squared_log_error, squared=False), "minimize"),
        "r2": (r2_score, "maximize"),
        "mse": (mean_squared_error, "minimize"),
        "mae": (mean_absolute_error, "minimize"),
    },
    "multilabel_classification": {
        "logloss": (log_loss, "minimize"),
    },
}


def get_metrics(task=None, metric=None):
    if task is None:
        return _METRICS
    if metric is None:
        return _METRICS[task]
    else:
        return _METRICS[task][metric]


def get_optuna_metric(task):
    return _OPTUNA_METRICS[task]



from dataclasses import dataclass, field

from biocore import DataHandler
from biocore.utils import get_kwargs

from biofit.processing import BaseProcessor, ProcessorConfig
from biofit.utils import logging

logger = logging.get_logger(__name__)


@dataclass
class FeatureSelectorConfig(ProcessorConfig):
    processor_type: str = field(default="feature_selection", init=False, repr=False)


class FeatureSelector(BaseProcessor):
    """Base class for feature selection processors."""

    config_class = FeatureSelectorConfig
    config: FeatureSelectorConfig

    def run(self, X, runner=None, fn_kwargs: dict = {}, **map_kwargs):
        fn_kwargs = self._prepare_runner(X, **fn_kwargs)
        if "func_type" in fn_kwargs and fn_kwargs["func_type"] != "_fit":
            input_format = fn_kwargs["in_format_kwargs"]["target_format"]
            fn_kwargs["out_format_kwargs"]["target_format"] = input_format
            runner = None
        return super().run(X, runner=runner, fn_kwargs=fn_kwargs, **map_kwargs)

    def _transform_any(self, X, selected_indices=None):
        if selected_indices is None:
            return X
        return DataHandler.select_columns(X, selected_indices)

    def _process_transform_batch_input(self, X, *fn_args, **fn_kwargs):
        func = fn_kwargs.get("fn", None)
        in_format_kwargs = fn_kwargs.get("in_format_kwargs", {})
        in_format_kwargs["input_columns"] = None
        input = DataHandler.to_format(X, **in_format_kwargs)
        _fn_kwargs = get_kwargs(fn_kwargs, func)

        selected_indices = fn_kwargs.get("selected_indices", None)
        unused_indices = fn_kwargs.get("unused_indices", None)
        keep_unused_columns = fn_kwargs.get("keep_unused_columns", None)
        if self.config._feature_names_out and DataHandler.supports_named_columns(input):
            feature_idx_out = DataHandler.get_column_indices(
                input, self.config._feature_names_out, raise_if_missing=False
            )
        elif self.config._feature_idx_out is not None:
            feature_idx_out = [
                selected_indices[i] for i in self.config._feature_idx_out
            ]
        else:
            raise ValueError(
                "FeatureSelectorConfig requires either _feature_idx_out, as well as "
                "_feature_idx_out when input format supports named columns."
            )
        if keep_unused_columns:
            feature_idx_out = list(sorted(feature_idx_out + unused_indices))
        _fn_kwargs["selected_indices"] = feature_idx_out
        return input, fn_args, _fn_kwargs

    def _process_fit_output(self, input, out):
        idx_out = self.config._feature_idx_out
        names_in = self.config.feature_names_in_
        names_out = self.config._feature_names_out
        if idx_out is not None:
            if names_in is not None and (
                names_out is None or len(idx_out) < len(names_out)
            ):
                names_out = [names_in[i] for i in idx_out]
        self.config._feature_names_out = names_out

        return super()._process_fit_output(input, out)

    def _process_transform_batch_output(self, input, out, **fn_kwargs):
        # do nothing
        return out

    def _process_transform_output(self, output, input, *args, **kwargs):
        unused_indices = kwargs.get("unused_indices", None)
        new_fingerprint = kwargs.get("fingerprint", None)
        selected_indices = self.config._feature_idx_out
        if DataHandler.supports_named_columns(input) and self.config._feature_names_out:
            input_cols = DataHandler.get_column_names(input)
            col_dict = dict(zip(input_cols, range(len(input_cols))))
            selected_indices = {
                col_dict[name]
                for name in self.config._feature_names_out
                if name in col_dict
            }

        before_filtering = len(kwargs.get("selected_indices", []))
        logger.info(
            f'"{self.__class__.__name__}": Selected {len(selected_indices)} out of '
            f"{before_filtering} input features"
        )

        if unused_indices:
            selected_indices = set(unused_indices).union(selected_indices)

        return DataHandler.select_columns(
            input, sorted(list(selected_indices)), new_fingerprint=new_fingerprint
        )



from pathlib import Path
from typing import List, Union

import numpy as np
from biocore import DataHandler
from sklearn.pipeline import Pipeline

from biofit.metrics import calculate_metrics, confusion_matrix, get_metrics
from biofit.processing import BaseProcessor
from biofit.train_eval_utils import (
    _get_data,
    get_model_info,
    preprocess,
    save_confusion_matrix,
    save_metrics,
    save_predictions,
)


def _predict(
    models,
    x_eval,
    y_eval=None,
    preprocessors=None,
    use_proba=True,
    task=None,
    label_names=None,
    cache_dir=None,
):
    # model can be a list of models for each y_eval, if multi-label
    # classification/regression is used
    if not isinstance(models, list):
        models = [models]

    y_preds = []
    for i, model in enumerate(models):
        preprocessor = preprocessors[i] if preprocessors is not None else None
        if isinstance(model, Pipeline):
            if len(model.steps) > 1 and preprocessor is None:
                preprocessor = Pipeline([p for p in model.steps[:-1]])
            model = model.steps[-1][1]
        if preprocessor:
            x_eval, _, _, _ = preprocess(
                preprocessor, x_eval, cache_dir=cache_dir, transform_only=True
            )

        def fn(model, use_proba, x_test):
            extra_kwargs = {}
            if isinstance(model, BaseProcessor):
                extra_kwargs = {"cache_dir": cache_dir, "load_from_cache_file": False}

            if use_proba and hasattr(model, "predict_proba"):
                return model.predict_proba(x_test, **extra_kwargs)
            elif hasattr(model, "predict"):
                return model.predict(x_test, **extra_kwargs)
            else:
                return model.transform(x_test, **extra_kwargs)

        if not hasattr(model, "predict") and not hasattr(model, "predict_proba"):
            for m in model:
                y_pred = fn(m, use_proba, x_eval)
        elif hasattr(model, "steps"):
            for step in model.steps:
                y_pred = fn(
                    step[-1] if isinstance(step, tuple) else step, use_proba, x_eval
                )
        else:
            y_pred = fn(model, use_proba, x_eval)

        y_pred_dims = DataHandler.get_shape(y_pred)
        x_test_dims = DataHandler.get_shape(x_eval)
        if y_eval is not None:
            if task == "multiclass_classification":
                if label_names is None:
                    label_names = np.unique(y_eval).tolist()

                if len(y_pred_dims) == 1 and (
                    (y_pred_dims[0] % len(label_names)) != 0
                    or (y_pred_dims[0] // len(label_names)) != x_test_dims[0]
                ):
                    y_pred_ohe = np.zeros((x_test_dims[0], len(label_names)))
                    y_pred_ohe[np.arange(x_test_dims[0]), y_pred] = 1
                    y_pred = y_pred_ohe
                elif y_pred_dims[1] != len(label_names):
                    y_pred_ohe = np.zeros((y_pred_dims[0], len(label_names)))
                    y_pred_ohe[np.arange(y_pred_dims[0]), y_pred] = 1
                    y_pred = y_pred_ohe

            if task == "binary_classification":
                if len(y_pred_dims) == 1 and (
                    (y_pred_dims[0] % 2) != 0 or (y_pred_dims[0] // 2) != x_test_dims[0]
                ):
                    y_pred = DataHandler.concat([1 - y_pred, y_pred], axis=1)
                elif len(y_pred_dims) > 1 and y_pred_dims[1] != 2:
                    y_pred = DataHandler.concat([1 - y_pred, y_pred], axis=1)

        y_preds.append(y_pred)

    if len(y_preds) == 1:
        y_preds = y_preds[0]
    else:
        y_preds = DataHandler.concat(y_preds, axis=1)
    return y_preds


def _update_metrics(y_true, y_pred, labels, metrics, task, results):
    calc_results = calculate_metrics(metrics, y_true, y_pred, task, labels)
    results.update(calc_results)


def evaluate(
    model,
    data,
    target=None,
    label_names=None,
    input_columns: Union[List[str], str] = "auto",
    target_columns: Union[List[str], str] = "auto",
    preprocessors=None,
    task=None,
    metrics=None,
    use_proba=True,
    output_dir=None,
    cache_dir=None,
    config=None,
):
    """
    Evaluate the model or models on the test data with improved flexibility and error
    handling.

    Args:
        data (np.ndarray): Data to be used for testing the model.
        target (np.ndarray, *optional*): Labels corresponding to the test data.
        models (list): A list of models to evaluate.
        preprocessor (object, *optional*):
            Preprocessor to apply to the data before evaluation.
        metrics (dict):
            Metrics to calculate for evaluation, with each metric being a callable. if
            None, all metrics for the `task` will be calculated.
        use_proba (bool):
            Whether to use the predict_proba method of the model for predictions.
        save_indices (bool): Whether to save the indices of the test set to a file.
        output_dir (str): Directory to save the predictions table
        cache_dir (str):
            Directory to save the cache files. Defaults to f"{output_dir}/cache", if
            output_dir is provided.

    Returns:
        list: A list of dictionaries with metric results for each model.
    """

    x_eval, y_eval, _, _, _, _, input_columns, target_columns = _get_data(
        data=data,
        target=target,
        valid_data=None,
        valid_target=None,
        input_columns=input_columns,
        target_columns=target_columns,
        format=None,
        target_required=False,
    )

    model_info = get_model_info(model, task)
    if task is None:
        task = model_info.get("task", None)

    if (
        preprocessors is not None
        and not isinstance(preprocessors, list)
        and isinstance(model, list)
    ):
        preprocessors = [preprocessors] * len(model)

    if "classification" in task and label_names is None:
        label_names = DataHandler.to_list(DataHandler.unique(y_eval))

    if task and metrics is None and y_eval is not None:
        if task == "classification":
            if len(label_names) > 2:
                task = "multiclass_classification"
            else:
                task = "binary_classification"
        metrics = get_metrics(task)

    if (
        preprocessors is not None
        and not isinstance(preprocessors, list)
        and isinstance(model, list)
    ):
        preprocessors = [preprocessors] * len(model)

    y_preds = _predict(
        x_eval=x_eval,
        models=model,
        preprocessors=preprocessors,
        use_proba=use_proba,
    )
    y_preds = DataHandler.to_pandas(y_preds)

    results = None
    if y_eval is not None:
        results = calculate_metrics(
            metrics, DataHandler.to_pandas(y_eval), y_preds, task, label_names
        )

    y_preds = y_preds.sort_index()
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        y_preds = DataHandler.to_pandas(y_preds)
        labs = model_info.get("class_names", None) or label_names
        if results is not None:
            save_metrics(output_dir, results)

            cm = confusion_matrix(y_eval, y_preds, labels=labs)
            save_confusion_matrix(output_dir, cm, label_names)
        save_predictions(
            output_dir,
            y_preds,
            data,
            y_eval,
            label_names=label_names,
            target_columns=target_columns,
        )

    return y_preds, results


def predict(
    model,
    data,
    input_columns: Union[List[str], str] = "auto",
    preprocessors=None,
    use_proba=True,
    output_dir=None,
    cache_dir=None,
    config=None,
):
    """
    Evaluate the model or models on the test data with improved flexibility and error
    handling.

    Args:
        data (np.ndarray): Data to be used for testing the model.
        target (np.ndarray, *optional*): Labels corresponding to the test data.
        models (list): A list of models to evaluate.
        preprocessor (object, *optional*):
            Preprocessor to apply to the data before evaluation.
        use_proba (bool):
            Whether to use the predict_proba method of the model for predictions.
        output_dir (str): Directory to save the predictions table
        cache_dir (str):
            Directory to save the cache files. Defaults to f"{output_dir}/cache", if
            output_dir is provided.
        config (dict, *optional*):
            Placeholder for additional configuration options. Currently not used.

    Returns:
        list: A list of dictionaries with metric results for each model.
    """

    model_info = get_model_info(model, task=None)
    x_eval, _, _, _, _, _, input_columns, _ = _get_data(
        data=data,
        input_columns=input_columns,
        target_columns=None,
        format=None,
        target_required=False,
    )

    if (
        preprocessors is not None
        and not isinstance(preprocessors, list)
        and isinstance(model, list)
    ):
        preprocessors = [preprocessors] * len(model)

    y_preds = _predict(
        x_eval=x_eval,
        models=model,
        preprocessors=preprocessors,
        use_proba=use_proba,
    )
    y_preds = DataHandler.to_pandas(y_preds)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        y_pred = DataHandler.to_pandas(y_preds)
        save_predictions(
            output_dir,
            y_pred,
            data,
            label_names=model_info.get("class_names", None),
        )

    return y_preds
